# Orca
#
# Copyright 2010 Joanmarie Diggs.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-lines
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-locals
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments

"""Utilities for providing app/toolkit-specific information about objects and events."""

from __future__ import annotations

import gi

gi.require_version("Atspi", "2.0")
from typing import TYPE_CHECKING

from gi.repository import Atspi

from . import (
    ax_event_synthesizer,
    debug,
    focus_manager,
    input_event_manager,
    messages,
    object_properties,
    presentation_manager,
    speech_presenter,
    spellcheck_presenter,
    table_navigator,
)
from .ax_object import AXObject
from .ax_table import AXTable
from .ax_text import AXText
from .ax_utilities import AXUtilities
from .ax_utilities_text import TextUnit

if TYPE_CHECKING:
    from .ax_utilities_text import CaretSetReason


class Utilities:
    """Utilities for providing app/toolkit-specific information about objects and events."""

    ZERO_WIDTH_NO_BREAK_SPACE = "\ufeff"

    def __init__(self, script):
        self._script = script

        # TODO - JD: We actually need to keep track of a root object and its descendant
        # caret context, just like we do in the web script. And it all should move into
        # the focus manager. In the meantime, just get things working for the active app.
        self._caret_context = (None, -1)

    def node_level(self, obj: Atspi.Accessible) -> int:
        """Returns the node level of the specified tree item."""

        if not AXUtilities.is_tree_or_tree_table_descendant(obj):
            return -1

        attrs = AXObject.get_attributes_dict(obj)
        if "level" in attrs:
            # ARIA levels are 1-based.
            return int(attrs.get("level", 0)) - 1

        nodes = []
        node = obj
        while node and (targets := AXUtilities.get_is_node_child_of(node)):
            node = targets[0]
            nodes.append(node)

        return len(nodes) - 1

    def child_nodes(self, obj: Atspi.Accessible) -> list[Atspi.Accessible]:
        """Gets all of the objects that have RELATION_NODE_CHILD_OF pointing to this cell."""

        if not AXUtilities.is_expanded(obj):
            return []

        table = AXUtilities.get_table(obj)
        if table is None:
            return []

        nodes = AXUtilities.get_is_node_parent_of(obj)
        tokens = ["SCRIPT UTILITIES:", len(nodes), "child nodes for", obj, "via node-parent-of"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        if nodes:
            return nodes

        # Candidates will be in the rows beneath the current row. Only check in the current column
        # and stop checking as soon as the node level of a candidate is equal or less than our
        # current level.
        row, col = AXTable.get_cell_coordinates(obj, prefer_attribute=False)
        level = self.node_level(obj)

        for i in range(row + 1, AXTable.get_row_count(table, prefer_attribute=False)):
            cell = AXTable.get_cell_at(table, i, col)
            targets = AXUtilities.get_is_node_child_of(cell)
            if not targets:
                continue

            node_of = targets[0]
            if obj == node_of:
                nodes.append(cell)
            elif self.node_level(node_of) <= level:
                break

        tokens = ["SCRIPT UTILITIES:", len(nodes), "child nodes for", obj, "via node-child-of"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return nodes

    def frame_and_dialog(
        self,
        obj: Atspi.Accessible | None = None,
    ) -> list[Atspi.Accessible | None]:
        """Returns the frame and (possibly) the dialog containing obj."""

        results = [None, None]

        obj = obj or focus_manager.get_manager().get_locus_of_focus()
        if not obj:
            msg = "SCRIPT UTILITIES: frame_and_dialog() called without valid object"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return results

        top_level = self.top_level_object(obj)
        if top_level is None:
            tokens = ["SCRIPT UTILITIES: could not find top-level object for", obj]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return results

        dialog_roles = [Atspi.Role.DIALOG, Atspi.Role.FILE_CHOOSER, Atspi.Role.ALERT]
        role = AXObject.get_role(top_level)
        if role in dialog_roles:
            results[1] = top_level
        else:
            if role in [Atspi.Role.FRAME, Atspi.Role.WINDOW]:
                results[0] = top_level

            def is_dialog(x):
                return AXObject.get_role(x) in dialog_roles

            results[1] = AXUtilities.find_ancestor_inclusive(obj, is_dialog)

        tokens = ["SCRIPT UTILITIES:", obj, "is in frame", results[0], "and dialog", results[1]]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return results

    def grab_focus_when_setting_caret(self, obj: Atspi.Accessible) -> bool:
        """Returns true if we should grab focus when setting the caret."""

        return AXUtilities.is_focusable(obj)

    def grab_focus_before_routing(self, obj: Atspi.Accessible) -> bool:
        """Returns true if we should grab focus before routing the cursor."""

        return (
            AXUtilities.is_combo_box(obj)
            and obj != focus_manager.get_manager().get_locus_of_focus()
        )

    def in_find_container(self, obj: Atspi.Accessible | None = None) -> bool:
        """Returns True if obj is in a find-in-page container."""

        if obj is None:
            obj = focus_manager.get_manager().get_locus_of_focus()

        if not AXUtilities.is_entry(obj):
            return False

        return AXUtilities.is_tool_bar_descendant(obj)

    def get_find_results_count(self, _root: Atspi.Accessible | None = None) -> str:
        """Returns a string description of the number of find-in-page results in root."""

        return ""

    def is_document(self, obj: Atspi.Accessible) -> bool:
        """Returns True if obj is a document."""

        # TODO - JD: See if the web script logic can be included here and then it all moved
        # into AXUtilities.
        return AXUtilities.is_document(obj)

    def in_document_content(self, obj: Atspi.Accessible | None = None) -> bool:
        """Returns True if obj (or the locus of focus) is in document content."""

        obj = obj or focus_manager.get_manager().get_locus_of_focus()
        return self.get_document_for_object(obj) is not None

    def active_document(self) -> Atspi.Accessible | None:
        """Returns the active document."""

        window = focus_manager.get_manager().get_active_window()
        documents = [
            obj
            for obj in AXUtilities.get_embeds(window)
            if self.is_document(obj) and AXUtilities.is_showing(obj)
        ]
        if len(documents) == 1:
            tokens = ["SCRIPT UTILITIES: Active document (via embeds):", documents[0]]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return documents[0]

        document = self.get_top_level_document_for_object(
            focus_manager.get_manager().get_locus_of_focus(),
        )
        tokens = ["SCRIPT UTILITIES: Active document (via locus of focus):", document]
        return document

    def is_top_level_document(self, obj: Atspi.Accessible) -> bool:
        """Returns true if obj is a top-level document."""

        return self.is_document(obj) and not AXUtilities.find_ancestor(obj, self.is_document)

    def get_top_level_document_for_object(self, obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the top-level document containing obj."""

        return AXUtilities.find_outermost_ancestor_inclusive(obj, self.is_document)

    def get_document_for_object(self, obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the nearest document ancestor of obj, or obj if it is a document."""

        # TODO - JD: Replace callers of this function with the logic below.
        return AXUtilities.find_ancestor_inclusive(obj, self.is_document)

    def should_read_full_row(
        self,
        obj: Atspi.Accessible,
        previous_object: Atspi.Accessible | None = None,
    ) -> bool:
        """Returns True if we should present the full row in speech."""

        if (
            focus_manager.get_manager().in_say_all()
            or table_navigator.get_navigator().last_input_event_was_navigation_command()
            or input_event_manager.get_manager().last_event_was_tab_navigation()
            or not AXUtilities.cell_row_changed(obj, previous_object)
        ):
            return False

        table = AXUtilities.get_table(obj)
        if table is None:
            return False

        manager = speech_presenter.get_presenter()
        if not self.get_document_for_object(table):
            return manager.get_speak_row_in_gui_table()
        if AXUtilities.is_spreadsheet_table(table):
            return manager.get_speak_row_in_spreadsheet()
        return manager.get_speak_row_in_document_table()

    def get_notification_content(self, obj: Atspi.Accessible) -> str:
        """Returns a string containing the content of the notification obj."""

        if not AXUtilities.is_notification(obj):
            return ""

        tokens = []
        name = AXObject.get_name(obj)
        if name:
            tokens.append(name)
        text = AXUtilities.expand_eocs(obj)
        if text and text not in tokens:
            tokens.append(text)
        else:
            labels = " ".join(
                AXText.get_all_text(x) or AXObject.get_name(x)
                for x in self.unrelated_labels(obj, False, 1)
            )
            if labels and labels not in tokens:
                tokens.append(labels)

        description = AXObject.get_description(obj)
        if description and description not in tokens:
            tokens.append(description)

        return " ".join(tokens)

    def is_link(self, obj: Atspi.Accessible) -> bool:
        """Returns True if obj is a link."""

        return AXUtilities.is_link(obj)

    def _is_top_level_object(self, obj: Atspi.Accessible) -> bool:
        return (
            AXObject.get_role(obj) in AXUtilities.get_top_level_roles()
            and AXObject.get_role(AXObject.get_parent(obj)) == Atspi.Role.APPLICATION
        )

    def top_level_object(
        self,
        obj: Atspi.Accessible,
        use_fallback_search: bool = False,
    ) -> Atspi.Accessible | None:
        """Returns the top-level object (frame, dialog ...) containing obj."""

        rv = AXUtilities.find_ancestor_inclusive(obj, self._is_top_level_object)
        tokens = ["SCRIPT UTILITIES:", rv, "is top-level object for:", obj]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if rv is None and use_fallback_search:
            msg = "SCRIPT UTILITIES: Attempting to find top-level object via fallback search"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            rv = AXUtilities.find_window_with_descendant(obj)

        return rv

    def top_level_object_is_active_and_current(self, obj: Atspi.Accessible | None = None) -> bool:
        """Returns true if the top-level object for obj/locus-of-focus is active and current."""

        # TODO - JD: Candidate for the focus manager.
        obj = obj or focus_manager.get_manager().get_locus_of_focus()
        top_level = self.top_level_object(obj)
        if not top_level:
            return False

        AXObject.clear_cache(top_level, False, "Ensuring we have the correct state.")
        if not AXUtilities.is_active(top_level) or AXUtilities.is_defunct(top_level):
            return False

        return top_level == focus_manager.get_manager().get_active_window()

    def unrelated_labels(
        self,
        root: Atspi.Accessible | None = None,
        only_showing: bool = True,
        minimum_words: int = 3,
    ) -> list[Atspi.Accessible]:
        """Returns a list of labels in root that lack a relationship."""

        if spellcheck_presenter.get_presenter().is_active():
            return []

        label_roles = [Atspi.Role.LABEL, Atspi.Role.STATIC]
        skip_roles = [
            Atspi.Role.BUTTON,
            Atspi.Role.COMBO_BOX,
            Atspi.Role.DOCUMENT_EMAIL,
            Atspi.Role.DOCUMENT_FRAME,
            Atspi.Role.DOCUMENT_PRESENTATION,
            Atspi.Role.DOCUMENT_SPREADSHEET,
            Atspi.Role.DOCUMENT_TEXT,
            Atspi.Role.DOCUMENT_WEB,
            Atspi.Role.FRAME,
            Atspi.Role.LIST_BOX,
            Atspi.Role.LIST,
            Atspi.Role.LIST_ITEM,
            Atspi.Role.MENU,
            Atspi.Role.MENU_BAR,
            Atspi.Role.SCROLL_PANE,
            Atspi.Role.SPLIT_PANE,
            Atspi.Role.TABLE,
            Atspi.Role.TOGGLE_BUTTON,
            Atspi.Role.TREE,
            Atspi.Role.TREE_TABLE,
            Atspi.Role.WINDOW,
        ]

        if AXObject.get_role(root) in skip_roles:
            return []

        root_is_notification = AXUtilities.is_notification(root)

        def _include(x):
            if not (x and AXObject.get_role(x) in label_roles):
                return False
            if not AXUtilities.object_is_unrelated(x):
                return False
            if not root_is_notification and AXUtilities.is_focusable(x):
                return False
            return not (only_showing and not AXUtilities.is_showing(x))

        def _exclude(x):
            if not x or AXObject.get_role(x) in skip_roles:
                return True
            return only_showing and not AXUtilities.is_showing(x)

        labels = AXUtilities.find_all_descendants(root, _include, _exclude)

        root_name = AXObject.get_name(root)

        # Eliminate things suspected to be labels for widgets
        labels_filtered = []
        for label in labels:
            name = AXObject.get_name(label) or AXText.get_all_text(label)
            if name and name in [root_name, AXObject.get_name(AXObject.get_parent(label))]:
                continue
            if len(name.split()) < minimum_words:
                continue
            if root_name.find(name) >= 0:
                continue
            labels_filtered.append(label)

        return AXUtilities.sort_objects_by_position(labels_filtered)

    def find_previous_object(
        self,
        obj: Atspi.Accessible,
        restrict_to: Atspi.Accessible | None = None,
    ) -> Atspi.Accessible | None:
        """Finds the object before this one."""

        if restrict_to is None:
            restrict_to = self.get_top_level_document_for_object(obj)
        return AXUtilities.find_previous_object(obj, restrict_to)

    def find_next_object(
        self,
        obj: Atspi.Accessible,
        restrict_to: Atspi.Accessible | None = None,
    ) -> Atspi.Accessible | None:
        """Finds the object after this one."""

        if restrict_to is None:
            restrict_to = self.get_top_level_document_for_object(obj)
        return AXUtilities.find_next_object(obj, restrict_to)

    def deleted_text(self, event: Atspi.Event) -> str:
        """Tries to determine the real deleted text for the given event. Because app bugs."""

        return event.any_data

    def inserted_text(self, event: Atspi.Event) -> str:
        """Tries to determine the real inserted text for the given event. Because app bugs."""

        if event.any_data:
            return event.any_data

        msg = "SCRIPT UTILITIES: Broken text insertion event"
        debug.print_message(debug.LEVEL_INFO, msg, True)

        if AXUtilities.is_password_text(event.source):
            string = AXText.get_all_text(event.source)
            if string:
                tokens = ["HACK: Returning last char in '", string, "'"]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return string[-1]

        msg = "FAIL: Unable to correct broken text insertion event"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return ""

    def clear_caret_context(
        self,
        document: Atspi.Accessible | None = None,  # pylint: disable=unused-argument
    ) -> None:
        """Clears the caret context."""

        # TODO - JD: This logic ultimately belongs in the focus manager.
        self._caret_context = (None, -1)

    def set_caret_context(
        self,
        obj: Atspi.Accessible | None = None,  # pylint: disable=unused-argument
        offset: int = -1,  # pylint: disable=unused-argument
        document: Atspi.Accessible | None = None,  # pylint: disable=unused-argument
    ) -> None:
        """Sets the caret context in document to (obj, offset)."""

        # TODO - JD: This logic ultimately belongs in the focus manager.
        self._caret_context = (obj, offset)

    def get_caret_context(
        self,
        document: Atspi.Accessible | None = None,  # pylint: disable=unused-argument
        search_if_needed: bool = True,  # pylint: disable=unused-argument
    ) -> tuple[Atspi.Accessible, int]:
        """Returns an (obj, offset) tuple representing the current location."""

        obj, offset = self._caret_context
        if obj is not None:
            tokens = ["SCRIPT UTILITIES: Returning cached caret context", obj, offset]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return obj, offset

        obj = focus_manager.get_manager().get_locus_of_focus()
        offset = AXText.get_caret_offset(obj)
        tokens = ["SCRIPT UTILITIES: Returning focus + caret offset", obj, offset]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return obj, offset

    def set_caret_position(
        self,
        obj: Atspi.Accessible,
        offset: int,
        document: Atspi.Accessible | None = None,  # pylint: disable=unused-argument
        *,
        reason: CaretSetReason,
    ) -> None:
        """Sets the locus of focus to obj and sets the caret position to offset."""

        manager = focus_manager.get_manager()
        manager.set_locus_of_focus(None, obj, False)
        if self.grab_focus_when_setting_caret(obj):
            AXObject.grab_focus(obj)

        if not reason.is_text_selection():
            # We cannot count on implementations clearing the selection for us when we set the
            # caret offset. Also, we should clear the selected text first.
            # https://bugs.documentfoundation.org/show_bug.cgi?id=167930
            AXUtilities.clear_all_selected_text(obj)
        self.set_caret_offset(obj, offset, reason=reason)

        # TODO - JD: The web script's set_caret_position() also sets the caret context.
        # Ensuring global structural navigation, caret navigation, browse mode, etc.
        # work means we should do the same here. Ultimately, however, we need to clearly
        # define and unite the whole offset + position + context + locus of focus +
        # last cursor position.
        manager.set_last_cursor_position(obj, offset)
        self.set_caret_context(obj, offset)

        scroll_to = max(0, min(offset, AXText.get_character_count(obj) - 1))
        ax_event_synthesizer.get_synthesizer().scroll_into_view(obj, scroll_to)

    def set_caret_offset(
        self, obj: Atspi.Accessible, offset: int, *, reason: CaretSetReason
    ) -> None:
        """Sets the caret offset via AtspiText."""

        # TODO - JD. Remove this function if the web override can be adjusted
        AXUtilities.set_caret_offset_with_reason(obj, offset, reason)

    def split_substring_by_language(
        self,
        obj: Atspi.Accessible,
        start: int,
        end: int,
    ) -> list[tuple[int, int, str, str, str]]:
        """Returns a list of (start, end, string, language, dialect) tuples."""

        rv: list[tuple[int, int, str, str, str]] = []
        all_substrings = self.get_language_and_dialect_from_text_attributes(obj, start, end)
        for start_offset, end_offset, language, dialect in all_substrings:
            if start >= end_offset:
                continue
            if end <= start_offset:
                break
            clamped_start = max(start, start_offset)
            clamped_end = min(end, end_offset)
            string = AXText.get_substring(obj, clamped_start, clamped_end)
            rv.append((clamped_start, clamped_end, string, language, dialect))

        return rv

    def get_language_and_dialect_for_substring(
        self,
        obj: Atspi.Accessible,
        start: int,
        end: int,
    ) -> tuple[str, str]:
        """Returns a (language, dialect) tuple. If multiple languages apply to
        the substring, language and dialect will be empty strings. Callers must
        do any preprocessing to avoid that condition."""

        all_substrings = self.get_language_and_dialect_from_text_attributes(obj, start, end)
        for start_offset, end_offset, language, dialect in all_substrings:
            if start_offset <= start and end_offset >= end:
                return language, dialect

        return "", ""

    def get_language_and_dialect_from_text_attributes(
        self,
        obj: Atspi.Accessible,
        start_offset: int = 0,
        end_offset: int = -1,
    ) -> list[tuple[int, int, str, str]]:
        """Returns a list of (start, end, language, dialect) tuples for obj."""

        rv: list[tuple[int, int, str, str]] = []
        attribute_set = AXUtilities.get_all_text_attributes(obj, start_offset, end_offset)
        last_language = last_dialect = ""
        for start, end, attrs in attribute_set:
            language = attrs.get("language", "")
            dialect = ""
            if "-" in language:
                language, dialect = language.split("-", 1)
            if rv and last_language == language and last_dialect == dialect:
                rv[-1] = rv[-1][0], end, language, dialect
            else:
                rv.append((start, end, language, dialect))
            last_language, last_dialect = language, dialect

        return rv

    def get_word_contents_at_offset(
        self,
        obj: Atspi.Accessible,
        offset: int = 0,
        use_cache: bool = True,  # pylint: disable=unused-argument
    ) -> list[tuple[Atspi.Accessible, int, int, str]]:
        """Returns a list of (obj, start, end, string) tuples for the word at offset."""

        text, start, end = AXText.get_word_at_offset(obj, offset)
        return [(obj, start, end, text)]

    def get_previous_line_contents(
        self,
        obj: Atspi.Accessible | None = None,  # pylint: disable=unused-argument
        offset: int = -1,  # pylint: disable=unused-argument
        layout_mode: bool | None = None,  # pylint: disable=unused-argument
        use_cache: bool = True,  # pylint: disable=unused-argument
    ) -> list[tuple[Atspi.Accessible, int, int, str]]:
        """Returns a list of (obj, start, end, string) tuples for the previous line."""

        if obj is None:
            obj, offset = self.get_caret_context()

        if offset == -1:
            offset = AXText.get_caret_offset(obj)

        text, start, end = AXUtilities.get_previous_line(obj, offset)
        if text:
            return [(obj, start, end, text)]

        _this_line, this_start, _this_end = AXText.get_line_at_offset(obj, offset)
        if not this_start:
            prev_obj, prev_offset = self.previous_context(obj, 0)
            return self.get_line_contents_at_offset(prev_obj, prev_offset)

        prev_obj, prev_offset = self.previous_context(obj, this_start)
        if prev_obj is not None:
            return self.get_line_contents_at_offset(prev_obj, prev_offset)

        return [(obj, 0, 0, "")]

    def get_line_contents_at_offset(
        self,
        obj: Atspi.Accessible,
        offset: int,
        layout_mode: bool | None = None,  # pylint: disable=unused-argument
        use_cache: bool = True,  # pylint: disable=unused-argument
    ) -> list[tuple[Atspi.Accessible, int, int, str]]:
        """Returns a list of (obj, start, end, string) tuples for the line at offset."""

        text, start, end = AXText.get_line_at_offset(obj, offset)
        return [(obj, start, end, text)]

    def get_next_line_contents(
        self,
        obj: Atspi.Accessible | None = None,
        offset: int = -1,
        layout_mode: bool | None = None,  # pylint: disable=unused-argument
        use_cache: bool = True,  # pylint: disable=unused-argument
    ) -> list[tuple[Atspi.Accessible, int, int, str]]:
        """Returns a list of (obj, start, end, string) tuples for the next line."""

        if obj is None:
            obj, offset = self.get_caret_context()

        if offset == -1:
            offset = AXText.get_caret_offset(obj)

        text, start, end = AXUtilities.get_next_line(obj, offset)
        if text:
            return [(obj, start, end, text)]

        _this_line, _this_start, this_end = AXText.get_line_at_offset(obj, offset)
        if this_end == AXText.get_character_count(obj):
            if offset < this_end:
                last_char = AXText.get_character_at_offset(obj, this_end - 1)[0]
                if not last_char or last_char in "\r\n":
                    return self.get_line_contents_at_offset(obj, this_end)

            next_obj, next_offset = self.next_context(obj, this_end)
            return self.get_line_contents_at_offset(next_obj, next_offset)

        next_obj, next_offset = self.next_context(obj, max(this_end, offset))
        if next_obj is not None:
            return self.get_line_contents_at_offset(next_obj, next_offset)

        return [(obj, 0, 0, "")]

    def get_object_contents_at_offset(
        self,
        obj: Atspi.Accessible,
        offset: int = 0,  # pylint: disable=unused-argument
        use_cache: bool = True,  # pylint: disable=unused-argument
    ) -> list[tuple[Atspi.Accessible, int, int, str]]:
        """Returns a list of (obj, start, end, string) tuples for the object at offset."""

        text = AXText.get_all_text(obj)
        return [(obj, 0, len(text), text)]

    def get_sentence_contents_at_offset(
        self,
        obj: Atspi.Accessible,
        offset: int,
        use_cache: bool = True,  # pylint: disable=unused-argument
    ) -> list[tuple[Atspi.Accessible, int, int, str]]:
        """Returns the sentence contents for the specified offset."""

        text, start, end = AXText.get_sentence_at_offset(obj, offset)
        return [(obj, start, end, text)]

    def previous_context(
        self,
        obj: Atspi.Accessible | None = None,
        offset: int = -1,
        skip_space: bool = False,
        restrict_to: Atspi.Accessible | None = None,
    ) -> tuple[Atspi.Accessible | None, int]:
        """Returns the previous viable/valid caret context given obj and offset."""

        if obj is None:
            obj, offset = self.get_caret_context()

        prev_offset = offset - 1
        if skip_space:
            while prev_offset >= 0:
                char = AXText.get_character_at_offset(obj, prev_offset)[0]
                if char and not char.isspace():
                    break
                prev_offset -= 1

        if prev_offset >= 0:
            return obj, prev_offset

        if prev_obj := self.find_previous_object(obj, restrict_to):
            if prev_obj != obj and not AXText.get_character_count(prev_obj) and not skip_space:
                return prev_obj, 0
            length = AXText.get_character_count(prev_obj)
            return self.previous_context(prev_obj, length, skip_space, restrict_to)

        return None, -1

    def next_context(
        self,
        obj: Atspi.Accessible | None = None,
        offset: int = -1,
        skip_space: bool = False,
        restrict_to: Atspi.Accessible | None = None,
    ) -> tuple[Atspi.Accessible | None, int]:
        """Returns the next viable/valid caret context given obj and offset."""

        if obj is None:
            obj, offset = self.get_caret_context()

        next_offset = offset + 1
        if skip_space:
            character_count = AXText.get_character_count(obj)
            while next_offset <= character_count:
                char = AXText.get_character_at_offset(obj, next_offset)[0]
                if char and not char.isspace():
                    break
                next_offset += 1

        if next_offset <= AXText.get_character_count(obj):
            return obj, next_offset

        if next_obj := self.find_next_object(obj, restrict_to):
            if next_obj != obj and not AXText.get_character_count(next_obj) and not skip_space:
                return next_obj, 0
            return self.next_context(next_obj, -1, skip_space, restrict_to)

        return None, -1

    def first_context(self, obj: Atspi.Accessible, offset: int) -> tuple[Atspi.Accessible, int]:
        """Returns the first viable/valid caret context given obj and offset."""

        if AXObject.supports_text(obj):
            return obj, offset

        descendant = AXUtilities.get_descendant_supporting_text(obj)
        if descendant is not None:
            return descendant, 0

        return obj, offset

    def filter_contents_for_presentation(
        self,
        contents: list[tuple[Atspi.Accessible, int, int, str]],
        infer_labels: bool = False,  # pylint: disable=unused-argument
    ) -> list[tuple[Atspi.Accessible, int, int, str]]:
        """Filters contents for presentation, removing objects that should not be included."""

        return contents

    def last_context(self, root: Atspi.Accessible) -> tuple[Atspi.Accessible, int]:
        """Returns the last viable/valid caret context in root."""

        offset = max(0, AXText.get_character_count(root) - 1)
        return root, offset

    def speak_selected_cell_range(self, _obj: Atspi.Accessible) -> bool:
        """Speaks the selected cell range in obj."""

        # TODO - JD: This doesn't belong here.
        return False

    def is_clickable_element(
        self,
        obj: Atspi.Accessible,  # pylint: disable=unused-argument
    ) -> bool:
        """Returns true if obj is a clickable element (in the web sense of that word)."""

        return False

    def has_long_desc(
        self,
        obj: Atspi.Accessible,  # pylint: disable=unused-argument
    ) -> bool:
        """Returns true if obj has a longdesc (deprecated web attribute)."""

        return False

    def has_meaningful_toggle_action(self, obj: Atspi.Accessible) -> bool:
        """Returns true if obj has a toggle action that is meaningful. Because app bugs."""

        return AXUtilities.has_action(obj, "toggle") or AXUtilities.has_action(
            obj,
            object_properties.ACTION_TOGGLE,
        )

    @staticmethod
    def _strip_newline_from_repeated_word(
        word: str,
        prev_word: str,
        start: int,
        end: int,
    ) -> tuple[int, int]:
        """Strips leading/trailing newline from a word that matches the previous word."""

        if "\n" not in word or word != prev_word:
            return start, end
        if word.startswith("\n"):
            return start + 1, end
        if word.endswith("\n"):
            return start, end - 1
        return start, end

    def get_word_at_offset_adjusted_for_navigation(
        self,
        obj: Atspi.Accessible,
        offset: int | None = None,
    ) -> tuple[str, int, int]:
        """Returns the word at offset, adjusted for native navigation commands."""

        if offset is None:
            offset = AXText.get_caret_offset(obj)

        word, start, end = AXText.get_word_at_offset(obj, offset)
        prev_obj, prev_offset = focus_manager.get_manager().get_penultimate_cursor_position()
        if prev_obj != obj:
            return word, start, end

        manager = input_event_manager.get_manager()
        was_previous_word_nav = manager.last_event_was_previous_word_navigation()
        was_next_word_nav = manager.last_event_was_next_word_navigation()

        # If we're in an ongoing series of native navigation-by-word commands, just present the
        # newly-traversed string.
        prev_word, prev_start, prev_end = AXText.get_word_at_offset(prev_obj, prev_offset)
        if AXUtilities.get_last_text_unit_spoken() == TextUnit.WORD:
            if was_previous_word_nav:
                start = offset
                end = prev_offset
            elif was_next_word_nav:
                start = prev_offset
                end = offset

            word = AXText.get_substring(obj, start, end)
            debug_string = word.replace("\n", "\\n")
            tokens = [
                "SCRIPT UTILITIES: Adjusted word at offset",
                offset,
                "for ongoing word nav is '",
                debug_string,
                "' (",
                start,
                "-",
                end,
                ")",
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return word, start, end

        # Otherwise, attempt some smarts so that the user winds up with the same presentation
        # they would get were this an ongoing series of native navigation-by-word commands.
        if was_previous_word_nav:
            # If we moved left via native nav, this should be the start of a native-navigation
            # word boundary, regardless of what ATK/AT-SPI2 tells us.
            start = offset

            # The ATK/AT-SPI2 word typically ends in a space; if the ending is neither a space,
            # nor an alphanumeric character, then suspect that character is a navigation boundary
            # where we would have landed before via the native previous word command.
            if not (word[-1].isspace() or word[-1].isalnum()):
                end -= 1

        elif was_next_word_nav:
            # If we moved right via native nav, this should be the end of a native-navigation
            # word boundary, regardless of what ATK/AT-SPI2 tells us.
            end = offset

            # This suggests we just moved to the end of the previous word.
            if word != prev_word and prev_start < offset <= prev_end:
                start = prev_start

            # If the character to the left of our present position is neither a space, nor
            # an alphanumeric character, then suspect that character is a navigation boundary
            # where we would have landed before via the native next word command.
            if offset > 0:
                last_char = AXText.get_substring(obj, offset - 1, offset)
                if not (last_char.isspace() or last_char.isalnum()):
                    start = offset - 1

        word = AXText.get_substring(obj, start, end)
        start, end = self._strip_newline_from_repeated_word(word, prev_word, start, end)
        word = AXText.get_substring(obj, start, end)
        debug_string = word.replace("\n", "\\n")
        tokens = [
            "SCRIPT UTILITIES: Adjusted word at offset",
            offset,
            "for new word nav is '",
            debug_string,
            "' (",
            start,
            "-",
            end,
            ")",
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return word, start, end

    def handle_container_selection_change(self, obj: Atspi.Accessible) -> bool:
        """Handles a change in a container that supports selection."""

        all_already_selected = AXUtilities.get_all_items_selected_state(obj)
        all_currently_selected = AXUtilities.all_items_selected(obj)
        AXUtilities.set_all_items_selected_state(obj, all_currently_selected)
        if all_already_selected and all_currently_selected:
            return True
        if input_event_manager.get_manager().last_event_was_select_all() and all_currently_selected:
            presentation_manager.get_manager().present_message(messages.CONTAINER_SELECTED_ALL)
            focus_manager.get_manager().set_locus_of_focus(None, obj, False)
            return True

        return False

    # pylint: disable=unused-argument
    def is_text_block_element(self, obj: Atspi.Accessible) -> bool:
        """Returns True if obj is a text block element like a paragraph or heading."""

        if not AXUtilities.is_text_block(obj, exclude_editable=True, exclude_focusable=True):
            return False

        return not self.has_name_and_action_and_no_useful_children(obj)

    def has_name_and_action_and_no_useful_children(self, obj: Atspi.Accessible) -> bool:
        """Returns True if obj has a name, supports action, and has no useful children."""

        return False

    def is_content_editable_with_embedded_objects(self, obj: Atspi.Accessible) -> bool:
        """Returns True if obj is content editable with embedded objects."""

        return False

    # pylint: enable=unused-argument



# Orca
#
# Copyright 2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Manages accessible text selection."""

from __future__ import annotations

import enum
from dataclasses import dataclass
from typing import TYPE_CHECKING

from . import ax_cache_manager, debug, input_event_manager
from .ax_object import AXObject
from .ax_text import AXText
from .ax_utilities import AXUtilities

if TYPE_CHECKING:
    from collections.abc import Hashable

    import gi

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi

    from . import input_event
    from .scripts import default

    SelectionPosition = tuple[Atspi.Accessible | None, int]
    SelectionBoundaries = tuple[SelectionPosition, SelectionPosition]


@dataclass(frozen=True)
class _TextSelectionEndpoint:
    """An accessible object and offset forming one endpoint of a text selection."""

    accessible_object: Atspi.Accessible
    offset: int


@dataclass(frozen=True)
class _TextSelectionSnapshot:
    """Selection state captured before an Orca text-selection command."""

    selection_container: Atspi.Accessible | None
    focus: _TextSelectionEndpoint
    focus_object_selection_start: int
    focus_object_selection_end: int
    anchor: _TextSelectionEndpoint


@dataclass(frozen=True, kw_only=True)
class _PendingSelectionChange:
    """A requested selection change awaiting its reported result."""

    container: Atspi.Accessible
    anchor: _TextSelectionEndpoint
    focus: _TextSelectionEndpoint


@dataclass(frozen=True)
class _TextSelectionResult:
    """Result of an Orca text-selection command."""

    succeeded: bool
    objects: tuple[Atspi.Accessible, ...]
    anchor: _TextSelectionEndpoint | None
    pending_change: _PendingSelectionChange | None = None


class SelectionChangeState(enum.Enum):
    """Describes how a reported text-selection change should be handled."""

    NOT_ORCA = enum.auto()
    UNPRESENTABLE = enum.auto()
    PRESENTABLE = enum.auto()


@dataclass(frozen=True, kw_only=True)
class TextSelectionCommand:
    """An Orca command which changed text selection."""

    _event: input_event.InputEvent
    _objects: tuple[Atspi.Accessible, ...]
    _selection_container: Atspi.Accessible | None
    _selection_anchor: _TextSelectionEndpoint | None
    _selection_focus: _TextSelectionEndpoint | None
    _pending_change: _PendingSelectionChange | None
    _should_notify_user: bool

    def get_input_event(self) -> input_event.InputEvent:
        """Returns the input event which triggered this selection command."""

        return self._event

    def get_objects(self) -> tuple[Atspi.Accessible, ...]:
        """Returns the objects associated with this selection command."""

        return self._objects

    def get_selection_container(self) -> Atspi.Accessible | None:
        """Returns the container associated with this selection command."""

        return self._selection_container

    def get_selection_anchor(self) -> _TextSelectionEndpoint | None:
        """Returns the anchor associated with this selection command."""

        return self._selection_anchor

    def get_selection_focus(self) -> _TextSelectionEndpoint | None:
        """Returns the focus associated with this selection command."""

        return self._selection_focus

    def get_pending_change(self) -> _PendingSelectionChange | None:
        """Returns the selection change awaiting its reported result."""

        return self._pending_change

    def should_notify_user(self) -> bool:
        """Returns True if this selection command should be presented."""

        return self._should_notify_user


class TextSelectionManager:
    """Provides high-level text-selection operations independent of navigation modality."""

    SELECTION_BOUNDARIES = "TextSelectionManager.boundaries"

    def __init__(self) -> None:
        self._last_selection_command: TextSelectionCommand | None = None
        self._pending_page_change: tuple[TextSelectionCommand, int] | None = None
        manager = ax_cache_manager.get_manager()
        manager.register_cache(
            self,
            self.SELECTION_BOUNDARIES,
            lifetime=ax_cache_manager.Lifetime.OWNER,
            clear_on_demand=ax_cache_manager.ClearPolicy.PRESERVE,
            clear_interval_seconds=None,
        )
        self._selection_boundaries = manager.get_cache(
            self,
            self.SELECTION_BOUNDARIES,
        )

    def get_current_selection_command(
        self,
        obj: Atspi.Accessible | None = None,
    ) -> TextSelectionCommand | None:
        """Returns the Orca selection command associated with the current input event."""

        command = self._get_current_selection_command()
        if command is None:
            return None
        if obj is not None and not self._command_applies_to_object(command, obj):
            return None
        return command

    def _get_current_selection_command(self) -> TextSelectionCommand | None:
        """Returns the selection command associated with the current input event."""

        command = self._last_selection_command
        if command is None:
            return None

        event = command.get_input_event()
        manager = input_event_manager.get_manager()
        result = manager.last_event_equals_or_is_release_for_event(event)
        string = event.as_single_line_string()
        tokens = ["TEXT SELECTION MANAGER: Last selection event (", string, ") is current:", result]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return command if result else None

    def get_selection_command_for_object(
        self,
        obj: Atspi.Accessible,
    ) -> TextSelectionCommand | None:
        """Returns the Orca command associated with a reported selection change."""

        command = self._last_selection_command
        if command is None or not self._command_applies_to_object(command, obj):
            return None
        if self._get_current_selection_command() is command:
            return command
        if command.get_pending_change() is None:
            self._clear_selection_command()
            return None
        return command

    def is_selection_change_from_selection_command(
        self,
        obj: Atspi.Accessible | None,
    ) -> bool:
        """Returns True if a selection command caused the reported change in obj."""

        if input_event_manager.get_manager().last_event_was_caret_selection():
            return True
        return bool(obj is not None and self.get_current_selection_command(obj) is not None)

    def defer_page_change_for_current_selection(self, page_number: int) -> bool:
        """Defers a page change associated with the current selection command."""

        command = self.get_current_selection_command()
        if command is None:
            self._pending_page_change = None
            return False

        self._pending_page_change = command, page_number
        tokens = ["TEXT SELECTION MANAGER: Deferred page", page_number, "for selection command."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return True

    def take_deferred_page_change_for_selection(
        self,
        obj: Atspi.Accessible,
    ) -> int | None:
        """Returns and clears the page change associated with obj's selection command."""

        pending = self._pending_page_change
        if pending is None:
            return None

        self._pending_page_change = None
        command, page_number = pending
        if self.get_selection_command_for_object(obj) is not command:
            return None
        return page_number

    def _clear_selection_command(self) -> None:
        """Clears the retained selection command and its pending page change."""

        self._last_selection_command = None
        self._pending_page_change = None

    @staticmethod
    def _command_applies_to_object(
        command: TextSelectionCommand,
        obj: Atspi.Accessible,
    ) -> bool:
        """Returns True if command applies to obj."""

        if obj in command.get_objects():
            return True

        container = command.get_selection_container()
        return bool(container is not None and AXUtilities.is_ancestor(obj, container, True))

    @staticmethod
    def _get_selected_text_with_eocs_expanded(obj: Atspi.Accessible) -> str:
        """Returns the selected text with embedded object characters expanded."""

        string, start_offset, end_offset = AXUtilities.get_selected_text(obj)
        if not string:
            return ""

        return AXUtilities.expand_eocs_in_range(
            obj,
            start_offset,
            obj,
            end_offset,
            include_start=True,
            include_end=False,
        )

    def get_all_selected_text(
        self,
        script: default.Script,
        obj: Atspi.Accessible,
    ) -> str:
        """Returns all text in the logical selection containing obj."""

        if AXUtilities.is_spreadsheet_cell(obj):
            return self._get_selected_text_with_eocs_expanded(obj)

        document = (
            obj
            if AXUtilities.is_document(obj)
            else AXUtilities.find_ancestor(obj, AXUtilities.is_document)
        )
        if document is not None:
            success, strings = AXUtilities.get_document_selected_texts(document)
            if success:
                return " ".join(strings)

            start, end = AXUtilities.get_document_text_selection_endpoints(None, document)
            start_obj, start_offset = start
            end_obj, end_offset = end
            if start_obj is not None and end_obj is not None:
                expanded = AXUtilities.expand_eocs_in_range(
                    start_obj,
                    start_offset,
                    end_obj,
                    end_offset,
                    include_start=True,
                    include_end=False,
                )
                if expanded:
                    return expanded

        strings = []
        current = self._get_selected_text_with_eocs_expanded(obj)
        if current:
            strings.append(current)

        preceding_strings: list[str] = []
        previous_obj = script.utilities.find_previous_object(obj)
        while previous_obj is not None:
            selection = self._get_selected_text_with_eocs_expanded(previous_obj)
            if not selection:
                break
            preceding_strings.insert(0, selection)
            previous_obj = script.utilities.find_previous_object(previous_obj)

        next_obj = script.utilities.find_next_object(obj)
        while next_obj is not None:
            selection = self._get_selected_text_with_eocs_expanded(next_obj)
            if not selection:
                break
            strings.append(selection)
            next_obj = script.utilities.find_next_object(next_obj)

        return " ".join(preceding_strings + strings)

    def _get_selection_anchor_from_previous_command(
        self,
        selection_container: Atspi.Accessible | None,
    ) -> _TextSelectionEndpoint | None:
        """Returns the selection anchor saved by the previous Orca command."""

        command = self._last_selection_command
        anchor = command.get_selection_anchor() if command is not None else None
        if (
            command is None
            or command.get_selection_container() != selection_container
            or anchor is None
        ):
            return None
        if not AXObject.is_valid(anchor.accessible_object):
            msg = "TEXT SELECTION MANAGER: Discarding invalid retained selection anchor."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return None

        return anchor

    def _clear_cached_selection_state(
        self,
        selection_root: Atspi.Accessible,
    ) -> SelectionBoundaries | None:
        """Clears and returns cached selection boundaries for selection_root."""

        key = ax_cache_manager.get_object_key(selection_root)
        boundaries_cache = self._selection_boundaries
        if boundaries_cache is None:
            return None

        boundaries = boundaries_cache.get(key, None)
        boundaries_cache.discard(key)
        if boundaries is None:
            return None
        tokens = [
            "TEXT SELECTION MANAGER: Cleared cached selection state for",
            selection_root,
            "Boundaries:",
            boundaries,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return boundaries

    def _get_selection_change_state(
        self,
        container: Atspi.Accessible | None,
        selection: SelectionBoundaries,
    ) -> tuple[SelectionChangeState, bool]:
        """Returns the state and whether stable selection data should be updated."""

        command = self._last_selection_command
        if command is None:
            return SelectionChangeState.NOT_ORCA, True

        pending_change = command.get_pending_change()
        if pending_change is None or pending_change.container != container:
            return SelectionChangeState.NOT_ORCA, True

        result_is_requested = self._selection_matches_pending_change(selection, pending_change)

        if self._get_current_selection_command() is command:
            if result_is_requested:
                return SelectionChangeState.PRESENTABLE, True
            return SelectionChangeState.UNPRESENTABLE, False

        self._clear_selection_command()
        if result_is_requested:
            msg = "TEXT SELECTION MANAGER: Ignoring change from previous selection command."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return SelectionChangeState.UNPRESENTABLE, True
        return SelectionChangeState.NOT_ORCA, True

    @staticmethod
    def _selection_matches_pending_change(
        selection: SelectionBoundaries,
        pending_change: _PendingSelectionChange,
    ) -> bool:
        """Returns True if selection matches the requested pending change."""

        anchor = pending_change.anchor
        focus = pending_change.focus
        comparison = AXUtilities.compare_text_positions(
            anchor.accessible_object,
            anchor.offset,
            focus.accessible_object,
            focus.offset,
        )
        start, end = selection
        if comparison == 0:
            return start[0] is None and end[0] is None
        if start[0] is None or end[0] is None:
            return False

        requested_start = anchor if comparison < 0 else focus
        requested_end = focus if comparison < 0 else anchor
        return AXUtilities.text_selection_positions_are_equivalent(
            start[0], start[1], requested_start.accessible_object, requested_start.offset
        ) and AXUtilities.text_selection_positions_are_equivalent(
            end[0], end[1], requested_end.accessible_object, requested_end.offset
        )

    @staticmethod
    def _get_document_and_selection_root(
        obj: Atspi.Accessible,
    ) -> tuple[Atspi.Accessible | None, Atspi.Accessible]:
        """Returns the document and selection root containing obj."""

        document = AXUtilities.find_outermost_ancestor_inclusive(obj, AXUtilities.is_document)
        selection_root = document or AXUtilities.get_text_selection_container(obj)
        return document, selection_root

    def has_known_selection(self, obj: Atspi.Accessible) -> bool:
        """Returns True if obj or its selection root is known to have selected text."""

        if AXUtilities.has_selected_text(obj):
            return True

        _document, selection_root = self._get_document_and_selection_root(obj)
        if selection_root is None:
            return False

        return self._has_cached_selection(ax_cache_manager.get_object_key(selection_root))

    def _has_cached_selection(self, key: Hashable) -> bool:
        """Returns True if the cached boundaries for key describe selected text."""

        start, end = self._get_cached_selection(key)
        return start[0] is not None or end[0] is not None

    @staticmethod
    def _get_text_selection_endpoints(
        anchor: _TextSelectionEndpoint | None,
        focus: _TextSelectionEndpoint | None,
    ) -> SelectionBoundaries:
        """Returns the ordered endpoints, or empty endpoints when no selection exists."""

        if anchor is None or focus is None:
            return (None, -1), (None, -1)
        anchor_position = anchor.accessible_object, anchor.offset
        focus_position = focus.accessible_object, focus.offset
        if anchor_position == focus_position:
            return (None, -1), (None, -1)
        if AXUtilities.compare_text_positions(*anchor_position, *focus_position) > 0:
            return focus_position, anchor_position
        return anchor_position, focus_position

    def get_known_text_selection_endpoints(
        self,
        selection_root: Atspi.Accessible,
    ) -> SelectionBoundaries:
        """Returns selection endpoints already known for selection_root without querying it."""

        command = self._last_selection_command
        if command is not None and command.get_selection_container() == selection_root:
            return self._get_text_selection_endpoints(
                command.get_selection_anchor(),
                command.get_selection_focus(),
            )

        return self._get_cached_selection(ax_cache_manager.get_object_key(selection_root))

    def _get_cached_selection(self, key: Hashable) -> SelectionBoundaries:
        """Returns the cached selection boundaries for key."""

        if self._selection_boundaries is None:
            return (None, -1), (None, -1)
        return self._selection_boundaries.get(key, ((None, -1), (None, -1)))

    @staticmethod
    def _update_selected_text_caches(
        old_selection: SelectionBoundaries,
        selection: SelectionBoundaries,
        event_source: Atspi.Accessible | None = None,
    ) -> None:
        """Updates selected-text caches for objects in both selections."""

        old_start, old_end = old_selection
        start, end = selection
        reported_elements = AXUtilities.get_text_selection_elements(
            old_start[0],
            old_end[0],
        ) + AXUtilities.get_text_selection_elements(start[0], end[0])
        if event_source is not None:
            reported_elements.append(event_source)
        elements: list[Atspi.Accessible] = []
        for element in reported_elements:
            if element not in elements:
                elements.append(element)
        for element in elements:
            AXUtilities.update_cached_selected_text(element)

    def _store_selection_change(
        self,
        selection_root: Atspi.Accessible,
        key: Hashable,
        state: SelectionChangeState,
        selection: SelectionBoundaries,
    ) -> None:
        """Caches a reported selection change and updates related state."""

        if self._selection_boundaries is not None:
            self._selection_boundaries.put(key, selection)
        start, end = selection
        if state == SelectionChangeState.PRESENTABLE and start[0] is None and end[0] is None:
            self._clear_cached_selection_state(selection_root)

    def _get_selection_endpoints_for_change(
        self,
        document: Atspi.Accessible | None,
        selection_root: Atspi.Accessible,
    ) -> SelectionBoundaries:
        """Returns the current selection endpoints."""

        success, endpoints = AXUtilities.get_text_selection_endpoints_from_document_interface(
            document
        )
        if success:
            return endpoints

        command = self._last_selection_command
        if command is not None and self._get_current_selection_command() is command:
            pending_change = command.get_pending_change()
            if pending_change is not None and pending_change.container == selection_root:
                return self._get_text_selection_endpoints(
                    pending_change.anchor,
                    pending_change.focus,
                )

        return AXUtilities.get_document_text_selection_endpoints(None, selection_root)

    def update_selection_state(
        self,
        obj: Atspi.Accessible,
    ) -> tuple[SelectionChangeState, SelectionBoundaries, SelectionBoundaries]:
        """Updates selection state for obj and returns its classification and boundaries."""

        document, selection_root = self._get_document_and_selection_root(obj)
        key = ax_cache_manager.get_object_key(selection_root)
        old_selection = self._get_cached_selection(key)
        old_start, old_end = old_selection
        selection = self._get_selection_endpoints_for_change(document, selection_root)
        start, end = selection
        tokens = [
            "TEXT SELECTION MANAGER: Updating text selection state for",
            obj,
            "Selection root:",
            selection_root,
            "Old boundaries:",
            old_start,
            old_end,
            "New boundaries:",
            start,
            end,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        state, update_state = self._get_selection_change_state(
            selection_root,
            selection,
        )
        if start[0] is None and end[0] is None and AXUtilities.has_selected_text(selection_root):
            msg = "TEXT SELECTION MANAGER: Ignoring indeterminate selection boundaries."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            state = SelectionChangeState.UNPRESENTABLE
            update_state = False

        if update_state:
            self._store_selection_change(
                selection_root,
                key,
                state,
                selection,
            )
        if state == SelectionChangeState.UNPRESENTABLE:
            self._update_selected_text_caches(old_selection, selection, obj)
        return state, old_selection, selection

    def update_state_for_unpresentable_selection_change(
        self,
        obj: Atspi.Accessible,
    ) -> None:
        """Updates selection state and caches without presenting the reported change."""

        state, old_selection, selection = self.update_selection_state(obj)
        if state == SelectionChangeState.UNPRESENTABLE:
            return
        self._update_selected_text_caches(old_selection, selection, obj)

    def set_text_selection(
        self,
        selection_container: Atspi.Accessible | None,
        source_object: Atspi.Accessible,
        source_offset: int,
        destination_object: Atspi.Accessible,
        destination_offset: int,
        destination_context_object: Atspi.Accessible,
        *,
        selection_forward: bool,
        event: input_event.InputEvent | None,
        notify_user: bool,
    ) -> bool:
        """Sets text selection from the source position to the destination position."""

        snapshot = self._take_snapshot(
            selection_container,
            source_object,
            source_offset,
        )
        return self._apply_selection(
            snapshot,
            destination_object,
            destination_offset,
            destination_context_object,
            selection_forward=selection_forward,
            event=event,
            notify_user=notify_user,
        )

    def _take_snapshot(
        self,
        selection_container: Atspi.Accessible | None,
        focus_object: Atspi.Accessible,
        focus_offset: int,
    ) -> _TextSelectionSnapshot:
        """Captures state needed to apply an Orca text-selection command."""

        _selected_string, selection_start, selection_end = AXUtilities.get_selected_text(
            focus_object
        )
        focus = _TextSelectionEndpoint(focus_object, focus_offset)
        anchor_offset_in_focus = AXUtilities.get_selection_anchor_offset(
            focus.accessible_object,
            focus.offset,
            selection_start,
            selection_end,
        )
        anchor = self._get_selection_anchor(
            selection_container,
            focus,
            anchor_offset_in_focus,
            selection_start != selection_end,
        )

        return _TextSelectionSnapshot(
            selection_container,
            focus,
            selection_start,
            selection_end,
            anchor,
        )

    def _get_selection_anchor(
        self,
        selection_container: Atspi.Accessible | None,
        focus: _TextSelectionEndpoint,
        anchor_offset_in_focus: int,
        focus_object_has_selection: bool,
    ) -> _TextSelectionEndpoint:
        """Returns the fixed anchor for the selection."""

        previous_anchor = self._get_selection_anchor_from_previous_command(selection_container)
        if previous_anchor is not None:
            return previous_anchor

        default_anchor = _TextSelectionEndpoint(focus.accessible_object, anchor_offset_in_focus)
        if selection_container is None or not (
            focus_object_has_selection or AXUtilities.has_selected_text(selection_container)
        ):
            return default_anchor

        start, end = AXUtilities.get_document_text_selection_endpoints(
            selection_container,
            selection_container,
        )
        start_object, start_offset = start
        end_object, end_offset = end
        if start_object is None or end_object is None:
            return default_anchor
        if AXUtilities.text_selection_positions_are_equivalent(
            focus.accessible_object,
            focus.offset,
            start_object,
            start_offset,
        ):
            return _TextSelectionEndpoint(end_object, end_offset)
        if AXUtilities.text_selection_positions_are_equivalent(
            focus.accessible_object,
            focus.offset,
            end_object,
            end_offset,
        ):
            return _TextSelectionEndpoint(start_object, start_offset)
        return default_anchor

    @staticmethod
    def _get_selection_range_for_source_object(
        snapshot: _TextSelectionSnapshot,
        selection_forward: bool,
        character_count: int,
    ) -> tuple[int, int]:
        """Returns the range to select in the object being left."""

        anchor_offset_in_focus = TextSelectionManager._get_anchor_offset_in_focus(snapshot)
        if selection_forward:
            return anchor_offset_in_focus, character_count
        return 0, anchor_offset_in_focus

    @staticmethod
    def _get_anchor_offset_in_focus(snapshot: _TextSelectionSnapshot) -> int:
        """Returns the selection anchor mapped into snapshot's focus object."""

        if snapshot.anchor.accessible_object == snapshot.focus.accessible_object:
            return snapshot.anchor.offset
        return AXUtilities.get_selection_anchor_offset(
            snapshot.focus.accessible_object,
            snapshot.focus.offset,
            snapshot.focus_object_selection_start,
            snapshot.focus_object_selection_end,
        )

    @staticmethod
    def _get_selection_range_for_destination_object(
        current_selection_start: int,
        current_selection_end: int,
        new_focus_offset: int,
        selection_forward: bool,
        character_count: int,
    ) -> tuple[int, int]:
        """Returns the range to select in the object being entered."""

        selection_exists = current_selection_start != current_selection_end
        if selection_exists and selection_forward:
            return new_focus_offset, current_selection_end
        if selection_exists:
            return current_selection_start, new_focus_offset
        if selection_forward:
            return 0, new_focus_offset
        return new_focus_offset, character_count

    @staticmethod
    def _set_text_object_selection(
        obj: Atspi.Accessible,
        start: int,
        end: int,
    ) -> bool:
        """Sets or clears the selection in obj for the requested range."""

        if start == end:
            AXUtilities.clear_all_selected_text(obj)
            return True

        result = AXUtilities.set_selected_text(obj, start, end)
        tokens = ["TEXT SELECTION MANAGER: Set selection:", result, obj, (start, end)]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    def _apply_selection(
        self,
        snapshot: _TextSelectionSnapshot,
        new_focus_object: Atspi.Accessible,
        new_focus_offset: int,
        new_caret_context_object: Atspi.Accessible,
        *,
        selection_forward: bool,
        event: input_event.InputEvent | None,
        notify_user: bool,
    ) -> bool:
        """Applies the selection represented by snapshot and the new focus."""

        new_focus = _TextSelectionEndpoint(new_focus_object, new_focus_offset)
        tokens = [
            "TEXT SELECTION MANAGER: Selection anchor is",
            snapshot.anchor.accessible_object,
            snapshot.anchor.offset,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        result = self._apply_selection_with_document_interface(
            snapshot,
            new_focus,
            new_caret_context_object,
        )
        if result is None and new_focus.accessible_object == snapshot.focus.accessible_object:
            result = self._apply_selection_within_text_object(
                snapshot,
                new_focus,
                new_caret_context_object,
            )
        elif result is None:
            result = self._apply_selection_across_text_objects(
                snapshot,
                new_focus,
                new_caret_context_object,
                selection_forward,
            )

        if not result.succeeded:
            msg = "TEXT SELECTION MANAGER: Selection command failed."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        self._record_selection_command(
            event,
            result.objects,
            notify_user,
            selection_container=snapshot.selection_container,
            selection_anchor=result.anchor,
            selection_focus=new_focus if result.anchor is not None else None,
            pending_change=result.pending_change,
        )
        return True

    @staticmethod
    def _apply_selection_with_document_interface(
        snapshot: _TextSelectionSnapshot,
        new_focus: _TextSelectionEndpoint,
        new_caret_context_object: Atspi.Accessible,
    ) -> _TextSelectionResult | None:
        """Applies selection through the document interface when available."""

        selection_container = snapshot.selection_container
        if selection_container is None:
            return None

        selection_was_set = AXUtilities.set_document_text_selection_endpoints(
            selection_container,
            snapshot.anchor.accessible_object,
            snapshot.anchor.offset,
            new_focus.accessible_object,
            new_focus.offset,
        )
        if not selection_was_set:
            return None

        selection_exists = (
            snapshot.anchor.accessible_object,
            snapshot.anchor.offset,
        ) != (
            new_focus.accessible_object,
            new_focus.offset,
        )
        pending_change = _PendingSelectionChange(
            container=selection_container,
            anchor=snapshot.anchor,
            focus=new_focus,
        )
        return _TextSelectionResult(
            True,
            (new_caret_context_object,),
            snapshot.anchor if selection_exists else None,
            pending_change,
        )

    def _apply_selection_within_text_object(
        self,
        snapshot: _TextSelectionSnapshot,
        new_focus: _TextSelectionEndpoint,
        new_caret_context_object: Atspi.Accessible,
    ) -> _TextSelectionResult:
        """Applies selection within one object through AtspiText."""

        text_object = snapshot.focus.accessible_object
        anchor_offset_in_focus = self._get_anchor_offset_in_focus(snapshot)
        selection_container = snapshot.selection_container
        if selection_container is not None and snapshot.anchor.accessible_object != text_object:
            container_start, container_end = AXUtilities.get_document_text_selection_endpoints(
                selection_container,
                selection_container,
            )
            if container_start[0] == snapshot.anchor.accessible_object:
                anchor_offset_in_focus = 0
            elif container_end[0] == snapshot.anchor.accessible_object:
                anchor_offset_in_focus = AXText.get_character_count(text_object)
            tokens = [
                "TEXT SELECTION MANAGER: AtspiText fallback mapped container anchor to offset",
                anchor_offset_in_focus,
                "in",
                text_object,
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        selection_exists = anchor_offset_in_focus != new_focus.offset
        if selection_exists:
            start, end = sorted((anchor_offset_in_focus, new_focus.offset))
            succeeded = AXUtilities.set_selected_text(text_object, start, end)
        else:
            AXUtilities.clear_all_selected_text(text_object)
            succeeded = True

        return _TextSelectionResult(
            succeeded,
            (new_caret_context_object,),
            snapshot.anchor if selection_exists else None,
        )

    def _apply_selection_across_text_objects(
        self,
        snapshot: _TextSelectionSnapshot,
        new_focus: _TextSelectionEndpoint,
        new_caret_context_object: Atspi.Accessible,
        selection_forward: bool,
    ) -> _TextSelectionResult:
        """Applies a cross-object selection through the AtspiText fallback."""

        source_object = snapshot.focus.accessible_object
        source_range = self._get_selection_range_for_source_object(
            snapshot,
            selection_forward,
            AXText.get_character_count(source_object),
        )
        destination_object = new_focus.accessible_object
        _string, destination_start, destination_end = AXUtilities.get_selected_text(
            new_focus.accessible_object
        )
        destination_range = self._get_selection_range_for_destination_object(
            destination_start,
            destination_end,
            new_focus.offset,
            selection_forward,
            AXText.get_character_count(destination_object),
        )

        old_source_range = (
            snapshot.focus_object_selection_start,
            snapshot.focus_object_selection_end,
        )
        source_range_is_empty = source_range[0] == source_range[1]
        old_source_range_is_empty = old_source_range[0] == old_source_range[1]
        source_range_changed = source_range != old_source_range and not (
            source_range_is_empty and old_source_range_is_empty
        )
        if source_range_changed:
            source_succeeded = self._set_text_object_selection(
                source_object,
                *source_range,
            )
            if not source_succeeded:
                self._set_text_object_selection(
                    source_object,
                    *old_source_range,
                )
                msg = "TEXT SELECTION MANAGER: Attempted to restore source after failure."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return _TextSelectionResult(False, (), None)

        old_destination_range = destination_start, destination_end
        destination_range_is_empty = destination_range[0] == destination_range[1]
        if not self._set_text_object_selection(
            destination_object,
            *destination_range,
        ):
            self._set_text_object_selection(
                destination_object,
                *old_destination_range,
            )
            if source_range_changed:
                self._set_text_object_selection(
                    source_object,
                    *old_source_range,
                )
            msg = "TEXT SELECTION MANAGER: Attempted to restore selection after failure."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return _TextSelectionResult(False, (), None)

        selection_exists = not source_range_is_empty or not destination_range_is_empty
        return _TextSelectionResult(
            True,
            (source_object, new_caret_context_object),
            snapshot.anchor if selection_exists else None,
        )

    def clear_selection_for_navigation(
        self,
        root: Atspi.Accessible | None,
        destination_obj: Atspi.Accessible,
    ) -> list[Atspi.Accessible]:
        """Clears selected text before ordinary navigation and returns affected objects."""

        command = self._last_selection_command
        self._clear_selection_command()
        selection_root = root or destination_obj
        cached_selection = self._clear_cached_selection_state(selection_root)
        candidates: list[Atspi.Accessible] = []

        def add_candidate(obj: Atspi.Accessible | None) -> None:
            if obj is not None and obj not in candidates:
                candidates.append(obj)

        def add_selection(boundaries: SelectionBoundaries) -> None:
            start, end = boundaries
            for endpoint_obj, _offset in (start, end):
                if endpoint_obj is not None:
                    add_candidate(AXUtilities.get_text_selection_container(endpoint_obj))
            for element in AXUtilities.get_text_selection_elements(start[0], end[0]):
                add_candidate(element)

        if cached_selection is not None:
            add_selection(cached_selection)
        if command is not None:
            add_selection(
                self._get_text_selection_endpoints(
                    command.get_selection_anchor(),
                    command.get_selection_focus(),
                )
            )
            for obj in command.get_objects():
                add_candidate(obj)

        destination_has_selection = AXUtilities.has_selected_text(destination_obj)
        if destination_has_selection and cached_selection is None and root is not None:
            start, end = AXUtilities.get_document_text_selection_endpoints(root, root)
            add_selection((start, end))
        add_candidate(destination_obj)

        cleared_selection_objs: list[Atspi.Accessible] = []
        for obj in candidates:
            if not AXObject.supports_text(obj):
                continue
            has_selection = (
                destination_has_selection
                if obj == destination_obj
                else AXUtilities.has_selected_text(obj)
            )
            if not has_selection:
                continue
            cleared_selection_objs.append(obj)
            AXUtilities.clear_all_selected_text(obj)
            AXUtilities.update_cached_selected_text(obj)
        return cleared_selection_objs

    def _record_selection_command(
        self,
        event: input_event.InputEvent | None,
        objects: tuple[Atspi.Accessible, ...],
        notify_user: bool,
        *,
        selection_container: Atspi.Accessible | None,
        selection_anchor: _TextSelectionEndpoint | None,
        selection_focus: _TextSelectionEndpoint | None,
        pending_change: _PendingSelectionChange | None,
    ) -> None:
        """Records an Orca selection command triggered by event."""

        if event is None:
            return

        self._pending_page_change = None
        self._last_selection_command = TextSelectionCommand(
            _event=event,
            _objects=objects,
            _selection_container=selection_container,
            _selection_anchor=selection_anchor,
            _selection_focus=selection_focus,
            _pending_change=pending_change,
            _should_notify_user=notify_user,
        )


_manager = TextSelectionManager()


def get_manager() -> TextSelectionManager:
    """Returns the singleton text selection manager."""

    return _manager



# Orca
#
# Copyright 2024 Igalia, S.L.
# Copyright 2024 GNOME Foundation Inc.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-public-methods
# pylint: disable=too-many-lines
# pylint: disable=too-many-branches
# pylint: disable=too-many-return-statements

"""Wrapper for the Atspi.Text interface."""

from __future__ import annotations

import contextlib
import enum
import locale
import re

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi, GLib

from . import (
    ax_cache_manager,
    colornames,
    debug,
    language_utilities,
    messages,
    object_properties,
    text_attribute_names,
)
from .ax_object import AXObject
from .ax_utilities_role import AXUtilitiesRole
from .ax_utilities_state import AXUtilitiesState

_TEXT_ATTRIBUTE_ALIASES: dict[str, str] = {
    "font-style": "style",
    "font-weight": "weight",
    "font-size": "size",
    "font-family": "family-name",
    "text-decoration-line": "text-decoration",
}


class AXTextAttribute(enum.Enum):
    """Enum representing an accessible text attribute."""

    # Note: Anything added here should also have an entry in text_attribute_names.py.
    # The tuple is (non-localized name, enable by default).
    BG_COLOR = ("bg-color", True)
    BG_FULL_HEIGHT = ("bg-full-height", False)
    BG_STIPPLE = ("bg-stipple", False)
    DIRECTION = ("direction", False)
    EDITABLE = ("editable", False)
    FAMILY_NAME = ("family-name", True)
    FG_COLOR = ("fg-color", True)
    FG_STIPPLE = ("fg-stipple", False)
    FONT_EFFECT = ("font-effect", True)
    INDENT = ("indent", True)
    INVALID = ("invalid", True)
    INVISIBLE = ("invisible", False)
    JUSTIFICATION = ("justification", True)
    LANGUAGE = ("language", False)
    LEFT_MARGIN = ("left-margin", False)
    LINE_HEIGHT = ("line-height", False)
    MARK = ("mark", True)
    PARAGRAPH_STYLE = ("paragraph-style", True)
    PIXELS_ABOVE_LINES = ("pixels-above-lines", False)
    PIXELS_BELOW_LINES = ("pixels-below-lines", False)
    PIXELS_INSIDE_WRAP = ("pixels-inside-wrap", False)
    RIGHT_MARGIN = ("right-margin", False)
    RISE = ("rise", False)
    SCALE = ("scale", False)
    SIZE = ("size", True)
    STRETCH = ("stretch", False)
    STRIKETHROUGH = ("strikethrough", True)
    STYLE = ("style", True)
    TEXT_DECORATION = ("text-decoration", True)
    TEXT_POSITION = ("text-position", False)
    TEXT_ROTATION = ("text-rotation", True)
    TEXT_SHADOW = ("text-shadow", False)
    UNDERLINE = ("underline", True)
    VARIANT = ("variant", False)
    VERTICAL_ALIGN = ("vertical-align", False)
    WEIGHT = ("weight", True)
    WRAP_MODE = ("wrap-mode", False)
    WRITING_MODE = ("writing-mode", False)

    @classmethod
    def from_string(cls, string: str) -> AXTextAttribute | None:
        """Returns the AXTextAttribute for the specified string."""

        canonical = _TEXT_ATTRIBUTE_ALIASES.get(string, string)
        for attribute in cls:
            if attribute.get_attribute_name() == canonical:
                return attribute

        return None

    @classmethod
    def from_localized_string(cls, string: str) -> AXTextAttribute | None:
        """Returns the AXTextAttribute for the specified localized string."""

        for attribute in cls:
            if attribute.get_localized_name() == string:
                return attribute

        return None

    def get_attribute_name(self) -> str:
        """Returns the non-localized name of the attribute."""

        return self.value[0]

    def get_value_from_attrs(self, attrs: dict[str, str]) -> str | None:
        """Returns the value for this attribute from an attrs dict, checking aliases."""

        key = self.value[0]
        value = attrs.get(key)
        if value is not None:
            return value
        for alias, canonical in _TEXT_ATTRIBUTE_ALIASES.items():
            if canonical == key:
                value = attrs.get(alias)
                if value is not None:
                    return value
        return None

    def get_localized_name(self) -> str:
        """Returns the localized name of the attribute."""

        name = self.value[0]
        return text_attribute_names.attribute_names.get(name, name)

    def get_localized_value(self, value) -> str:
        """Returns the localized value of the attribute."""

        if value is None:
            return ""

        if value.endswith("px"):
            value = value.split("px")[0]
            if locale.localeconv()["decimal_point"] in value:
                return messages.pixel_count(float(value))
            return messages.pixel_count(int(value))

        if self in [AXTextAttribute.BG_COLOR, AXTextAttribute.FG_COLOR]:
            return colornames.get_presentable_color_name(value)

        # TODO - JD: Is this still needed?
        value = value.replace("-moz", "")

        # TODO - JD: Are these still needed?
        if self == AXTextAttribute.JUSTIFICATION:
            value = value.replace("justify", "fill")
        elif self == AXTextAttribute.FAMILY_NAME:
            value = value.split(",")[0].strip().strip('"')

        return text_attribute_names.attribute_values.get(value, value)

    def should_present_by_default(self) -> bool:
        """Returns True if the attribute should be presented by default."""

        return self.value[1]

    def value_is_default(self, value) -> bool:
        """Returns True if value should be treated as the default value for this attribute."""

        null_values = ["0", "0mm", "0px", "none", "false", "normal", "", None]
        if value in null_values:
            return True

        if self == AXTextAttribute.SCALE:
            return float(value) == 1.0
        if self in (AXTextAttribute.TEXT_POSITION, AXTextAttribute.VERTICAL_ALIGN):
            return value == "baseline"
        if self == AXTextAttribute.WEIGHT:
            return value == "400"
        if self == AXTextAttribute.LANGUAGE:
            return value == language_utilities.get_current_language()

        return False

    def get_change_description(self, new_value: str | None) -> str:
        """Returns a localized description for announcing a change."""

        if self == AXTextAttribute.WEIGHT:
            if new_value and (
                new_value == "bold" or (new_value.isdigit() and int(new_value) > 400)
            ):
                return messages.BOLD
            return messages.TEXT_ATTRIBUTE_OFF % messages.BOLD

        if self == AXTextAttribute.STYLE:
            localized_italic = self.get_localized_value("italic")
            if new_value in ("italic", "oblique"):
                return self.get_localized_value(new_value)
            return messages.TEXT_ATTRIBUTE_OFF % localized_italic

        if self in (AXTextAttribute.UNDERLINE, AXTextAttribute.STRIKETHROUGH):
            if self.value_is_default(new_value):
                return messages.TEXT_ATTRIBUTE_OFF % self.get_localized_name()
            if new_value in ("single", "solid", "true"):
                return self.get_localized_name()
            return f"{self.get_localized_name()}: {self.get_localized_value(new_value)}"

        if self == AXTextAttribute.MARK:
            if self.value_is_default(new_value):
                return messages.TEXT_ATTRIBUTE_OFF % self.get_localized_name()
            return self.get_localized_name()

        if self == AXTextAttribute.INVALID:
            if new_value == "spelling":
                return messages.MISSPELLED
            if new_value == "grammar":
                return object_properties.STATE_INVALID_GRAMMAR_SPEECH
            return ""

        if self in (AXTextAttribute.TEXT_POSITION, AXTextAttribute.VERTICAL_ALIGN):
            if self.value_is_default(new_value):
                return messages.TEXT_ATTRIBUTE_OFF % self.get_localized_name()
            localized = self.get_localized_value(new_value)
            # TODO - JD: LO reports percentages instead of "super"/"sub".
            if localized == new_value and new_value and new_value.endswith("%"):
                value = new_value.removesuffix("%")
                with contextlib.suppress(ValueError):
                    localized = self.get_localized_value("super" if float(value) > 0 else "sub")
            return localized

        return f"{self.get_localized_name()}: {self.get_localized_value(new_value)}"


class AXText:
    """Wrapper for the Atspi.Text interface."""

    _CHARACTER_COUNT = "AXText.character-count"
    _ALL_TEXT = "AXText.all-text"
    _CACHE = ax_cache_manager.ScopedCache(
        (_CHARACTER_COUNT, _ALL_TEXT),
        ax_cache_manager.active_stable_tree_scope,
    )
    _ZERO_WIDTH_JOINER = "\u200d"
    _ZERO_WIDTH_NO_BREAK_SPACE = "\ufeff"
    _VARIATION_SELECTOR_RANGE = ("\ufe00", "\ufe0f")
    _EMOJI_MODIFIER_RANGE = ("\U0001f3fb", "\U0001f3ff")
    _REGIONAL_INDICATOR_RANGE = ("\U0001f1e6", "\U0001f1ff")
    # Widest span (in code points) a handled emoji cluster can occupy, including Gecko's
    # interleaved zero-width no-break spaces, with headroom. Bounds the text read for adjustment.
    _WHOLE_CHARACTER_MARGIN = 64

    @staticmethod
    def _extends_previous_character(text: str, offset: int) -> bool:
        """Returns True if text[offset] continues the previous character."""

        char = text[offset]
        if char == AXText._ZERO_WIDTH_NO_BREAK_SPACE:
            return True

        # Gecko interleaves zero-width no-break spaces, so look past them for the real neighbor.
        index = offset - 1
        while index >= 0 and text[index] == AXText._ZERO_WIDTH_NO_BREAK_SPACE:
            index -= 1
        if index < 0:
            return False
        previous = text[index]

        if AXText._ZERO_WIDTH_JOINER in (char, previous):
            return True
        vs_low, vs_high = AXText._VARIATION_SELECTOR_RANGE
        mod_low, mod_high = AXText._EMOJI_MODIFIER_RANGE
        if vs_low <= char <= vs_high or mod_low <= char <= mod_high:
            return True
        ri_low, ri_high = AXText._REGIONAL_INDICATOR_RANGE
        if ri_low <= char <= ri_high and ri_low <= previous <= ri_high:
            # Two regional indicators form one flag, so break only after a complete pair.
            run = 0
            while index >= 0:
                if text[index] == AXText._ZERO_WIDTH_NO_BREAK_SPACE:
                    index -= 1
                    continue
                if not ri_low <= text[index] <= ri_high:
                    break
                run += 1
                index -= 1
            return run % 2 == 1
        return False

    @staticmethod
    def _adjust_to_whole_characters(text: str, start: int, end: int) -> tuple[int, int]:
        """Extends start and end outward so neither falls inside a multi-code-point character."""

        length = len(text)
        start = max(0, min(start, length))
        end = max(start, min(end, length))
        while 0 < start < length and AXText._extends_previous_character(text, start):
            start -= 1
        while 0 < end < length and AXText._extends_previous_character(text, end):
            end += 1
        return start, end

    @staticmethod
    def _whole_character_content(
        obj: Atspi.Accessible, start: int, end: int
    ) -> tuple[str, int, int]:
        """Returns (content, start, end) grown to whole characters from a bounded text window."""

        length = AXText.get_character_count(obj)
        lo = max(0, start - AXText._WHOLE_CHARACTER_MARGIN)
        hi = min(length, end + AXText._WHOLE_CHARACTER_MARGIN)
        text = AXText.get_substring(obj, lo, hi)
        adjusted_start, adjusted_end = AXText._adjust_to_whole_characters(
            text, start - lo, end - lo
        )
        if (adjusted_start == 0 and lo > 0) or (adjusted_end == len(text) and hi < length):
            # The cluster reached the window edge; reread the full text to be safe.
            text = AXText.get_all_text(obj)
            adjusted_start, adjusted_end = AXText._adjust_to_whole_characters(text, start, end)
            return text[adjusted_start:adjusted_end], adjusted_start, adjusted_end
        return text[adjusted_start:adjusted_end], adjusted_start + lo, adjusted_end + lo

    @staticmethod
    def get_character_at_offset(
        obj: Atspi.Accessible,
        offset: int | None = None,
        ensure_whole_characters: bool = False,
    ) -> tuple[str, int, int]:
        """Returns the (character, start, end) for the current or specified offset."""

        length = AXText.get_character_count(obj)
        if not length:
            return "", 0, 0

        if offset is None:
            offset = AXText.get_caret_offset(obj)

        if not 0 <= offset <= length:
            tokens = ["WARNING: Offset", offset, "is not valid. No character can be provided."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return "", 0, 0

        try:
            result = Atspi.Text.get_string_at_offset(obj, offset, Atspi.TextGranularity.CHAR)
        except GLib.GError as error:
            tokens = ["AXText: Exception in get_character_at_offset:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return "", 0, 0

        if result is None:
            tokens = ["AXText: get_string_at_offset (char) failed for", obj, "at", offset, "."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return "", 0, 0

        debug_string = result.content.replace("\n", "\\n")
        tokens = [
            "AXText: Character at offset",
            offset,
            "in",
            obj,
            "'",
            debug_string,
            "' (",
            result.start_offset,
            "-",
            result.end_offset,
            ")",
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        content, start, end = result.content, result.start_offset, result.end_offset
        if ensure_whole_characters and start >= 0:
            if end < start:
                # Some engines return a broken boundary inside a multi-code-point
                # character (e.g. Chromium between a flag's regional indicators).
                start, end = offset, offset + 1
            content, start, end = AXText._whole_character_content(obj, start, end)
        return content, start, end

    @staticmethod
    def get_word_at_offset(
        obj: Atspi.Accessible,
        offset: int | None = None,
        ensure_whole_characters: bool = False,
    ) -> tuple[str, int, int]:
        """Returns the (word, start, end) for the current or specified offset."""

        length = AXText.get_character_count(obj)
        if not length:
            return "", 0, 0

        if offset is None:
            offset = AXText.get_caret_offset(obj)

        offset = min(max(0, offset), length - 1)
        try:
            result = Atspi.Text.get_string_at_offset(obj, offset, Atspi.TextGranularity.WORD)
        except GLib.GError as error:
            tokens = ["AXText: Exception in get_word_at_offset:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return "", 0, 0

        if result is None:
            tokens = ["AXText: get_string_at_offset (word) failed for", obj, "at", offset, "."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return "", 0, 0

        tokens = [
            "AXText: Word at offset",
            offset,
            "in",
            obj,
            "'",
            result.content,
            "' (",
            result.start_offset,
            "-",
            result.end_offset,
            ")",
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        content, start, end = result.content, result.start_offset, result.end_offset
        if ensure_whole_characters and start >= 0:
            if end < start:
                # The engine returned a broken word boundary inside a multi-code-point
                # character (e.g. Chromium inside a flag); recover the word from whitespace.
                # Rare and limited to small web objects, so reading the full text is fine.
                text = AXText.get_all_text(obj)
                start = offset
                while start > 0 and not text[start - 1].isspace():
                    start -= 1
                end = offset
                while end < len(text) and not text[end].isspace():
                    end += 1
                start, end = AXText._adjust_to_whole_characters(text, start, end)
                content = text[start:end]
            else:
                content, start, end = AXText._whole_character_content(obj, start, end)
        return content, start, end

    @staticmethod
    def get_line_at_offset(
        obj: Atspi.Accessible,
        offset: int | None = None,
    ) -> tuple[str, int, int]:
        """Returns the (line, start, end) for the current or specified offset."""

        length = AXText.get_character_count(obj)
        if not length:
            return "", 0, 0

        if offset is None:
            offset = AXText.get_caret_offset(obj)

        # Don't adjust the length in multiline text because we want to say "blank" at the end.
        # This may or may not be sufficient. GTK3 seems to give us the correct, empty line. But
        # (at least) Chromium does not. See comment below.
        if (
            not AXUtilitiesState.is_multi_line(obj)
            and not AXUtilitiesRole.is_paragraph(obj)
            and not AXUtilitiesRole.is_section(obj)
        ):
            offset = min(max(0, offset), length - 1)
        else:
            offset = max(0, offset)

        try:
            result = Atspi.Text.get_string_at_offset(obj, offset, Atspi.TextGranularity.LINE)
        except GLib.GError as error:
            tokens = ["AXText: Exception in get_line_at_offset:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return "", 0, 0

        if result is None:
            tokens = ["AXText: get_string_at_offset (line) failed for", obj, "at", offset, "."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return "", 0, 0

        # Try again, e.g. Chromium returns "", -1, -1.
        if result.start_offset == result.end_offset == -1 and offset == length:
            offset -= 1
            result = Atspi.Text.get_string_at_offset(obj, offset, Atspi.TextGranularity.LINE)

        if (
            debug.debugLevel <= debug.LEVEL_INFO
            and result.start_offset == result.end_offset == -1
            and 0 < offset < length
            and AXObject.get_toolkit_name(obj).startswith("chromium")
        ):
            tokens = [
                "ERROR: Chromium returned invalid line boundary at offset",
                offset,
                ". See https://crbug.com/536662914",
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        debug_string = result.content.replace("\n", "\\n")
        tokens = [
            "AXText: Line at offset",
            offset,
            "in",
            obj,
            "'",
            debug_string,
            "' (",
            result.start_offset,
            "-",
            result.end_offset,
            ")",
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if 0 <= offset < result.start_offset:
            offset -= 1
            tokens = ["ERROR: Start offset is greater than offset. Trying with offset", offset]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            result = Atspi.Text.get_string_at_offset(obj, offset, Atspi.TextGranularity.LINE)

            debug_string = result.content.replace("\n", "\\n")
            tokens = [
                "AXText: Line at offset",
                offset,
                "in",
                obj,
                "'",
                debug_string,
                "' (",
                result.start_offset,
                "-",
                result.end_offset,
                ")",
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        return result.content, result.start_offset, result.end_offset

    @staticmethod
    def _find_sentence_boundaries(text: str) -> list[int]:
        """Returns the offsets in text that should be treated as sentence beginnings."""

        if not text:
            return []

        boundaries = [0]
        pattern = r"[.!?]+(?=\s|\ufffc|$)|\n[^\S\n]*\n"
        for match in re.finditer(pattern, text):
            end_pos = match.end()
            # Skip whitespace to find start of next sentence. Do not skip embedded object
            # characters since they represent child objects that must be traversed.
            while end_pos < len(text) and text[end_pos].isspace():
                end_pos += 1
            # Only add boundary if we haven't reached the end and it's not a duplicate.
            if end_pos < len(text) and end_pos not in boundaries:
                boundaries.append(end_pos)

        if boundaries[-1] != len(text):
            boundaries.append(len(text))

        return boundaries

    @staticmethod
    def _get_sentence_at_offset_fallback(
        obj: Atspi.Accessible,
        offset: int | None = None,
    ) -> tuple[str, int, int]:
        """Fallback sentence detection for broken implementations."""

        if offset is None:
            offset = AXText.get_caret_offset(obj)

        text = AXText.get_all_text(obj)
        if not text or offset < 0 or offset >= len(text):
            return "", 0, 0

        fallback_text, fallback_start, fallback_end = text, 0, len(text)
        boundaries = AXText._find_sentence_boundaries(text)
        for i in range(len(boundaries) - 1):
            start, end = boundaries[i], boundaries[i + 1]
            if start <= offset < end:
                fallback_text, fallback_start, fallback_end = text[start:end], start, end
                break

        tokens = [
            "AXText: Fallback sentence in",
            obj,
            "at offset",
            offset,
            ": '",
            fallback_text,
            "' (",
            fallback_start,
            "-",
            fallback_end,
            ")",
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return fallback_text, fallback_start, fallback_end

    @staticmethod
    def get_sentence_at_offset(
        obj: Atspi.Accessible,
        offset: int | None = None,
    ) -> tuple[str, int, int]:
        """Returns the (sentence, start, end) for the current or specified offset."""

        length = AXText.get_character_count(obj)
        if not length:
            return "", 0, 0

        if offset is None:
            offset = AXText.get_caret_offset(obj)

        offset = min(max(0, offset), length - 1)
        try:
            result = Atspi.Text.get_string_at_offset(obj, offset, Atspi.TextGranularity.SENTENCE)
        except GLib.GError as error:
            tokens = ["AXText: Exception in get_sentence_at_offset:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return AXText._get_sentence_at_offset_fallback(obj, offset)

        if result is None:
            tokens = ["AXText: get_string_at_offset (sentence) failed for", obj, "at", offset, "."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return "", 0, 0

        if result.start_offset == result.end_offset == -1 or not result.content:
            return AXText._get_sentence_at_offset_fallback(obj, offset)

        if (
            result.start_offset == result.end_offset
            and result.start_offset in [0, -1]
            and not result.content
        ):
            return AXText._get_sentence_at_offset_fallback(obj, offset)

        if re.search(r"\n[^\S\n]*\n", result.content):
            fallback = AXText._get_sentence_at_offset_fallback(obj, offset)
            if result.start_offset <= fallback[1] <= offset < fallback[
                2
            ] <= result.end_offset and fallback[1:] != (result.start_offset, result.end_offset):
                return fallback

        tokens = [
            "AXText: Sentence at offset",
            offset,
            "in",
            obj,
            "'",
            result.content,
            "' (",
            result.start_offset,
            "-",
            result.end_offset,
            ")",
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result.content, result.start_offset, result.end_offset

    @staticmethod
    def get_paragraph_at_offset(
        obj: Atspi.Accessible,
        offset: int | None = None,
    ) -> tuple[str, int, int]:
        """Returns the (paragraph, start, end) for the current or specified offset."""

        length = AXText.get_character_count(obj)
        if not length:
            return "", 0, 0

        if offset is None:
            offset = AXText.get_caret_offset(obj)

        offset = min(max(0, offset), length - 1)
        try:
            result = Atspi.Text.get_string_at_offset(obj, offset, Atspi.TextGranularity.PARAGRAPH)
        except GLib.GError as error:
            tokens = ["AXText: Exception in get_paragraph_at_offset:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return "", 0, 0

        if result is None:
            tokens = ["AXText: get_string_at_offset (paragraph) failed for", obj, "at", offset, "."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return "", 0, 0

        tokens = [
            "AXText: Paragraph at offset",
            offset,
            "in",
            obj,
            "'",
            result.content,
            "' (",
            result.start_offset,
            "-",
            result.end_offset,
            ")",
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result.content, result.start_offset, result.end_offset

    @staticmethod
    def get_character_count(obj: Atspi.Accessible) -> int:
        """Returns the character count of obj."""

        if not AXObject.supports_text(obj):
            return 0

        key = ax_cache_manager.get_object_key(obj)
        cached = AXText._CACHE.get(AXText._CHARACTER_COUNT, key)
        if cached is not ax_cache_manager.MISSING:
            return cached

        try:
            count = Atspi.Text.get_character_count(obj)
        except GLib.GError as error:
            tokens = ["AXText: Exception in get_character_count:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return 0

        tokens = ["AXText:", obj, "reports", count, "characters."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        AXText._CACHE.set(AXText._CHARACTER_COUNT, key, count)
        return count

    @staticmethod
    def get_caret_offset(obj: Atspi.Accessible) -> int:
        """Returns the caret offset of obj."""

        if not AXObject.supports_text(obj):
            return -1

        try:
            offset = Atspi.Text.get_caret_offset(obj)
        except GLib.GError as error:
            tokens = ["AXText: Exception in get_caret_offset:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return -1

        tokens = ["AXText:", obj, "reports caret offset of", offset, "."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return offset

    @staticmethod
    def set_caret_offset(obj: Atspi.Accessible, offset: int) -> bool:
        """Returns False if we definitely failed to set the offset. True cannot be trusted."""

        if not AXObject.supports_text(obj):
            return False

        try:
            result = Atspi.Text.set_caret_offset(obj, offset)
        except GLib.GError as error:
            tokens = ["AXText: Exception in set_caret_offset:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        tokens = ["AXText: Reported result of setting offset to", offset, "in", obj, ":", result]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    @staticmethod
    def get_substring(obj: Atspi.Accessible, start_offset: int, end_offset: int) -> str:
        """Returns the text of obj within the specified offsets."""

        if not AXObject.supports_text(obj):
            return ""

        if end_offset == -1:
            end_offset = AXText.get_character_count(obj)

        try:
            result = Atspi.Text.get_text(obj, start_offset, end_offset)
        except GLib.GError as error:
            tokens = ["AXText: Exception in get_substring:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return ""

        debug_string = result.replace("\n", "\\n")
        tokens = [
            "AXText: Text of",
            obj,
            "(",
            start_offset,
            "-",
            end_offset,
            "): '",
            debug_string,
            "'",
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    @staticmethod
    def get_all_text(obj: Atspi.Accessible, length: int | None = None) -> str:
        """Returns the text content of obj."""

        key = ax_cache_manager.get_object_key(obj)
        cached = AXText._CACHE.get(AXText._ALL_TEXT, key)
        if cached is not ax_cache_manager.MISSING:
            return cached

        if length is None:
            length = AXText.get_character_count(obj)
        if not length:
            return ""

        try:
            result = Atspi.Text.get_text(obj, 0, length)
        except GLib.GError as error:
            tokens = ["AXText: Exception in get_all_text:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return ""

        if debug.debugLevel <= debug.LEVEL_INFO:
            words = result.split()
            if len(words) > 20:
                debug_string = f"{' '.join(words[:5])} ... {' '.join(words[-5:])}"
            else:
                debug_string = result

            debug_string = debug_string.replace("\n", "\\n")
            tokens = ["AXText: Text of", obj, "'", debug_string, "'"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        AXText._CACHE.set(AXText._ALL_TEXT, key, result)
        return result

    @staticmethod
    def get_n_selections(obj: Atspi.Accessible) -> int:
        """Returns the number of reported selected substrings in obj."""

        if not AXObject.supports_text(obj):
            return 0

        try:
            result = Atspi.Text.get_n_selections(obj)
        except GLib.GError as error:
            tokens = ["AXText: Exception in get_n_selections:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return 0

        tokens = ["AXText:", obj, "reports", result, "selection(s)."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    @staticmethod
    def remove_selection(obj: Atspi.Accessible, selection_number: int) -> None:
        """Attempts to remove the specified selection."""

        if not AXObject.supports_text(obj):
            return

        try:
            Atspi.Text.remove_selection(obj, selection_number)
        except GLib.GError as error:
            tokens = ["AXText: Exception in remove_selection:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return

    @staticmethod
    def get_selected_ranges(obj: Atspi.Accessible) -> list[tuple[int, int]]:
        """Returns a list of (start_offset, end_offset) tuples reflecting the selected text."""

        count = AXText.get_n_selections(obj)
        if not count:
            return []

        selections = []
        for i in range(count):
            try:
                result = Atspi.Text.get_selection(obj, i)
            except GLib.GError as error:
                tokens = ["AXText: Exception in get_selected_ranges:", error]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                break
            if 0 <= result.start_offset < result.end_offset:
                selections.append((result.start_offset, result.end_offset))

        tokens = ["AXText:", obj, "reports selected ranges:", selections]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return selections

    @staticmethod
    def add_new_selection(obj: Atspi.Accessible, start_offset: int, end_offset: int) -> bool:
        """Creates a new selection for the specified range in obj."""

        if not AXObject.supports_text(obj):
            return False

        try:
            result = Atspi.Text.add_selection(obj, start_offset, end_offset)
        except GLib.GError as error:
            tokens = ["AXText: Exception in add_new_selection:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        return result

    @staticmethod
    def update_existing_selection(
        obj: Atspi.Accessible,
        start_offset: int,
        end_offset: int,
        selection_number: int = 0,
    ) -> bool:
        """Modifies specified selection in obj to the specified range."""

        if not AXObject.supports_text(obj):
            return False

        try:
            result = Atspi.Text.set_selection(obj, selection_number, start_offset, end_offset)
        except GLib.GError as error:
            tokens = ["AXText: Exception in update_existing_selection:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        return result

    # TODO - JD: This should be converted to return AXTextAttribute values.
    @staticmethod
    def get_text_attributes_at_offset(
        obj: Atspi.Accessible,
        offset: int | None = None,
    ) -> tuple[dict[str, str], int, int]:
        """Returns a (dict, start, end) tuple for attributes at offset in obj."""

        if not AXObject.supports_text(obj):
            return {}, 0, 0

        if offset is None:
            offset = AXText.get_caret_offset(obj)

        try:
            result = Atspi.Text.get_attribute_run(obj, offset, include_defaults=True)
        except GLib.GError as error:
            tokens = ["AXText: Exception in get_text_attributes_at_offset:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return {}, 0, AXText.get_character_count(obj)

        if result is None or result[0] is None:
            tokens = ["AXText: get_attribute_run failed for", obj, "at offset", offset, "."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return {}, 0, AXText.get_character_count(obj)

        tokens = ["AXText: Attributes for", obj, "at offset", offset, ":", result]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        # Adjust for web browsers that report indentation and justification at object attributes
        # rather than text attributes.
        obj_attributes = AXObject.get_attributes_dict(obj, False)
        if not result[0].get("justification"):
            alternative = obj_attributes.get("text-align")
            if alternative:
                result[0]["justification"] = alternative
        if not result[0].get("indent"):
            alternative = obj_attributes.get("text-indent")
            if alternative:
                result[0]["indent"] = alternative

        return result[0] or {}, result[1] or 0, result[2] or AXText.get_character_count(obj)

    @staticmethod
    def get_offset_at_point(obj: Atspi.Accessible, x: int, y: int) -> int:
        """Returns the character offset in obj at the specified point."""

        if not AXObject.supports_text(obj):
            return -1

        try:
            offset = Atspi.Text.get_offset_at_point(obj, x, y, Atspi.CoordType.WINDOW)
        except GLib.GError as error:
            tokens = ["AXText: Exception in get_offset_at_point:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return -1

        tokens = ["AXText: Offset in", obj, "at", x, ",", y, "is", offset]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return offset

    @staticmethod
    def get_character_rect(obj: Atspi.Accessible, offset: int | None = None) -> Atspi.Rect:
        """Returns the Atspi rect of the character at the specified offset in obj."""

        if not AXObject.supports_text(obj):
            return Atspi.Rect()

        if offset is None:
            offset = AXText.get_caret_offset(obj)

        try:
            rect = Atspi.Text.get_character_extents(obj, offset, Atspi.CoordType.WINDOW)
        except GLib.GError as error:
            tokens = ["AXText: Exception in get_character_rect:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return Atspi.Rect()

        tokens = ["AXText: Offset", offset, "in", obj, "has rect", rect]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return rect

    @staticmethod
    def get_range_rect(obj: Atspi.Accessible, start: int, end: int) -> Atspi.Rect:
        """Returns the Atspi rect of the string at the specified range in obj."""

        if not AXObject.supports_text(obj):
            return Atspi.Rect()

        if end <= 0:
            end = AXText.get_character_count(obj)

        try:
            rect = Atspi.Text.get_range_extents(obj, start, end, Atspi.CoordType.WINDOW)
        except GLib.GError as error:
            tokens = ["AXText: Exception in get_range_rect:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return Atspi.Rect()

        tokens = ["AXText: Range", start, "-", end, "in", obj, "has rect", rect]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return rect

    @staticmethod
    def scroll_substring_to_point(
        obj: Atspi.Accessible,
        x: int,
        y: int,
        start_offset: int | None = None,
        end_offset: int | None = None,
    ) -> bool:
        """Attempts to scroll obj to the specified point."""

        length = AXText.get_character_count(obj)
        if not length:
            return False

        if start_offset is None:
            start_offset = 0
        if end_offset is None:
            end_offset = length - 1

        try:
            result = Atspi.Text.scroll_substring_to_point(
                obj,
                start_offset,
                end_offset,
                Atspi.CoordType.WINDOW,
                x,
                y,
            )
        except GLib.GError as error:
            tokens = ["AXText: Exception in scroll_substring_to_point:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        tokens = [
            "AXText: Scrolled",
            obj,
            "substring (",
            start_offset,
            "-",
            end_offset,
            ") to",
            x,
            ",",
            y,
            ":",
            result,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    @staticmethod
    def scroll_substring_to_location(
        obj: Atspi.Accessible,
        location: Atspi.ScrollType,
        start_offset: int | None = None,
        end_offset: int | None = None,
    ) -> bool:
        """Attempts to scroll the substring to the specified Atspi.ScrollType location."""

        length = AXText.get_character_count(obj)
        if not length:
            return False

        if start_offset is None:
            start_offset = 0
        if end_offset is None:
            end_offset = length - 1

        try:
            result = Atspi.Text.scroll_substring_to(obj, start_offset, end_offset, location)
        except GLib.GError as error:
            tokens = ["AXText: Exception in scroll_substring_to_location:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        tokens = [
            "AXText: Scrolled",
            obj,
            "substring (",
            start_offset,
            "-",
            end_offset,
            ") to",
            location,
            ":",
            result,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result



# Orca
#
# Copyright 2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Presents changes in text selection."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from . import (
    debug,
    document_presenter,
    input_event_manager,
    messages,
    presentation_manager,
    speech_presenter,
    text_selection_manager,
)
from .ax_object import AXObject
from .ax_text import AXText
from .ax_utilities import AXUtilities

if TYPE_CHECKING:
    import gi

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi

    from .scripts import default


class TextSelectionPresenter:
    """Presents changes in text selection."""

    def present_selected_text(
        self,
        script: default.Script,
        obj: Atspi.Accessible,
    ) -> bool:
        """Presents all text in the logical selection containing obj."""

        text = text_selection_manager.get_manager().get_all_selected_text(script, obj)
        if not text:
            presentation_manager.get_manager().speak_message(messages.NO_SELECTED_TEXT)
            return True

        manager = speech_presenter.get_presenter()
        indentation = manager.get_indentation_description(text, only_if_changed=False)
        text = manager.adjust_for_presentation(obj, text)
        message = messages.SELECTED_TEXT_IS % f"{indentation} {text}"
        presentation_manager.get_manager().speak_message(message)
        return True

    @staticmethod
    def present_selection_removed() -> None:
        """Presents that the current text selection was removed."""

        presentation_manager.get_manager().speak_message(messages.SELECTION_REMOVED)

    @staticmethod
    def _present_pending_page_change(obj: Atspi.Accessible) -> bool:
        """Presents a page change associated with obj's selection command."""

        page_number = text_selection_manager.get_manager().take_deferred_page_change_for_selection(
            obj
        )
        if page_number is None:
            return False
        presentation_manager.get_manager().present_message(messages.PAGE_NUMBER % page_number)
        return True

    def _compute_changes(
        self,
        script: default.Script,
        obj: Atspi.Accessible,
        old_string: str,
        old_start: int,
        old_end: int,
        new_string: str,
        new_start: int,
        new_end: int,
    ) -> tuple[list[list], bool]:
        """Returns the changes and whether a preceding child change was presented."""

        old_chars = set(range(old_start, old_end))
        new_chars = set(range(new_start, new_end))
        if not old_chars.union(new_chars):
            return [], False

        if old_chars and new_chars and not old_chars.intersection(new_chars):
            return (
                [
                    [old_start, old_end, messages.TEXT_UNSELECTED],
                    [new_start, new_end, messages.TEXT_SELECTED],
                ],
                False,
            )

        change = sorted(old_chars.symmetric_difference(new_chars))
        if not change:
            return [], False

        changes = []
        preceding_child_change_presented = False
        change_start, change_end = change[0], change[-1] + 1
        if old_chars < new_chars:
            changes.append([change_start, change_end, messages.TEXT_SELECTED])
            if old_string.endswith("\ufffc") and old_end == change_start:
                child = AXUtilities.find_child_at_offset(obj, old_end - 1)
                preceding_child_change_presented = self._handle_basic_change(
                    script,
                    child,
                    False,
                )
        else:
            changes.append([change_start, change_end, messages.TEXT_UNSELECTED])
            if new_string.endswith("\ufffc"):
                child = AXUtilities.find_child_at_offset(obj, new_end - 1)
                preceding_child_change_presented = self._handle_basic_change(
                    script,
                    child,
                    False,
                )
        return changes, preceding_child_change_presented

    def _present_changes(
        self,
        script: default.Script,
        obj: Atspi.Accessible,
        changes: list[list],
        speak_message: bool,
        preceding_child_change_presented: bool,
    ) -> None:
        """Presents the supplied selection changes."""

        speak_message = (
            speak_message and not speech_presenter.get_presenter().get_only_speak_displayed_text()
        )
        for start, end, message in changes:
            string = AXText.get_substring(obj, start, end)
            ends_with_child = string.endswith("\ufffc")
            effective_end = end - 1 if ends_with_child else end
            child = (
                AXUtilities.find_child_at_offset(obj, effective_end) if ends_with_child else None
            )
            tokens = [
                "TEXT SELECTION PRESENTER: Presenting change in",
                obj,
                "range",
                start,
                "-",
                end,
                "message='",
                message,
                "'",
                "ends with child:",
                ends_with_child,
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

            child_processed = False
            message_presented = False
            destination_child_change_presented = False
            if child is not None and string[:-1].isspace() and AXObject.supports_text(child):
                child_processed = True
                destination_child_change_presented = self._handle_basic_change(
                    script,
                    child,
                    speak_message,
                )
                if destination_child_change_presented:
                    tokens = [
                        "TEXT SELECTION PRESENTER: Suppressing whitespace before changed child",
                        child,
                    ]
                    debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                    continue

            if len(string) > 5000 and speak_message:
                if message == messages.TEXT_SELECTED:
                    presentation_manager.get_manager().speak_message(
                        messages.selected_character_count(len(string)),
                    )
                else:
                    presentation_manager.get_manager().speak_message(
                        messages.unselected_character_count(len(string)),
                    )
                message_presented = True
            else:
                script.say_phrase(obj, start, effective_end)
                if speak_message and (not ends_with_child or child_processed):
                    presentation_manager.get_manager().speak_message(message)
                    message_presented = True

            if ends_with_child:
                if child is None:
                    continue
                if AXObject.supports_text(child) and not child_processed:
                    destination_child_change_presented = self._handle_basic_change(
                        script,
                        child,
                        speak_message,
                    )
                else:
                    presentation_manager.get_manager().present_object(
                        script,
                        child,
                        generate_braille=False,
                    )
                    if speak_message:
                        presentation_manager.get_manager().speak_message(message)
                        message_presented = True

            if (
                speak_message
                and preceding_child_change_presented
                and not message_presented
                and not destination_child_change_presented
            ):
                presentation_manager.get_manager().speak_message(message)
            preceding_child_change_presented = False

    def _handle_basic_change(
        self,
        script: default.Script,
        obj: Atspi.Accessible,
        speak_message: bool,
    ) -> bool:
        """Handles a selection change in obj."""

        if (
            not AXObject.supports_text(obj)
            or input_event_manager.get_manager().last_event_was_cut()
        ):
            tokens = ["TEXT SELECTION PRESENTER: Ignoring basic change for", obj]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        old_string, old_start, old_end = AXUtilities.get_cached_selected_text(obj)
        AXUtilities.update_cached_selected_text(obj)
        new_string, new_start, new_end = AXUtilities.get_cached_selected_text(obj)
        tokens = [
            "TEXT SELECTION PRESENTER: Selection in",
            obj,
            "changed from",
            old_start,
            "-",
            old_end,
            "(",
            len(old_string),
            "chars)",
            "to",
            new_start,
            "-",
            new_end,
            "(",
            len(new_string),
            "chars)",
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if input_event_manager.get_manager().last_event_was_select_all() and new_string:
            if new_string != old_string:
                presentation_manager.get_manager().speak_message(messages.DOCUMENT_SELECTED_ALL)
            return True

        if (
            not text_selection_manager.get_manager().is_selection_change_from_selection_command(obj)
            and old_string
            and not new_string
        ):
            if speak_message:
                presentation_manager.get_manager().speak_message(messages.SELECTION_REMOVED)
            return False

        changes, preceding_child_change_presented = self._compute_changes(
            script,
            obj,
            old_string,
            old_start,
            old_end,
            new_string,
            new_start,
            new_end,
        )
        tokens = ["TEXT SELECTION PRESENTER: Computed changes for", obj, changes]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        if not changes:
            return False

        self._present_pending_page_change(obj)
        self._present_changes(
            script,
            obj,
            changes,
            speak_message,
            preceding_child_change_presented,
        )
        return True

    def _get_document_text_change(
        self,
        old_start: tuple[Atspi.Accessible | None, int],
        old_end: tuple[Atspi.Accessible | None, int],
        start: tuple[Atspi.Accessible | None, int],
        end: tuple[Atspi.Accessible | None, int],
    ) -> (
        tuple[
            tuple[Atspi.Accessible, int],
            tuple[Atspi.Accessible, int],
            bool,
            bool,
            str,
        ]
        | None
    ):
        """Returns the changed document range and its selection state."""

        old_start_obj, _old_start_offset = old_start
        old_end_obj, _old_end_offset = old_end
        start_obj, _start_offset = start
        end_obj, _end_offset = end
        if old_start_obj is None and old_end_obj is None:
            if start_obj is not None and end_obj is not None:
                return start, end, True, False, messages.TEXT_SELECTED
            return None
        if start_obj is None and end_obj is None:
            if old_start_obj is not None and old_end_obj is not None:
                return old_start, old_end, True, False, messages.TEXT_UNSELECTED
            return None

        if None in (old_start_obj, old_end_obj, start_obj, end_obj):
            return None
        starts_are_same = AXUtilities.compare_text_positions(*start, *old_start) == 0
        ends_are_same = AXUtilities.compare_text_positions(*end, *old_end) == 0
        if starts_are_same and not ends_are_same:
            comparison = AXUtilities.compare_text_positions(*old_end, *end)
            if comparison < 0:
                return old_end, end, True, False, messages.TEXT_SELECTED
            if comparison > 0:
                return end, old_end, True, False, messages.TEXT_UNSELECTED
        if ends_are_same and not starts_are_same:
            comparison = AXUtilities.compare_text_positions(*old_start, *start)
            if comparison > 0:
                return start, old_start, True, False, messages.TEXT_SELECTED
            if comparison < 0:
                return old_start, start, True, False, messages.TEXT_UNSELECTED
        return None

    def _present_document_text_change(
        self,
        script: default.Script,
        old_start: tuple[Atspi.Accessible | None, int],
        old_end: tuple[Atspi.Accessible | None, int],
        start: tuple[Atspi.Accessible | None, int],
        end: tuple[Atspi.Accessible | None, int],
        speak_message: bool,
    ) -> bool:
        """Presents a document text selection change as a single phrase."""

        change = self._get_document_text_change(old_start, old_end, start, end)
        if change is None:
            msg = "TEXT SELECTION PRESENTER: Could not identify changed document text range."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        range_start, range_end, include_start, include_end, message = change
        selection_obj = start[0] or old_start[0]
        if not text_selection_manager.get_manager().is_selection_change_from_selection_command(
            selection_obj
        ):
            selection_was_removed = start[0] is None and end[0] is None
            if not selection_was_removed:
                msg = "TEXT SELECTION PRESENTER: Change is not from a selection command."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return False

            msg = "TEXT SELECTION PRESENTER: Presenting selection removal."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            if speak_message:
                presentation_manager.get_manager().speak_message(messages.SELECTION_REMOVED)
            return True

        start_obj, start_offset = range_start
        end_obj, end_offset = range_end
        string = AXUtilities.expand_eocs_in_range(
            start_obj,
            start_offset,
            end_obj,
            end_offset,
            include_start=include_start,
            include_end=include_end,
        )
        tokens: list[Any] = [
            "TEXT SELECTION PRESENTER: Expanded changed document text range",
            range_start,
            range_end,
            "with endpoint inclusion",
            include_start,
            ",",
            include_end,
            "to '",
            string,
            "'",
            "message='",
            message,
            "'",
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        speak_message = (
            speak_message and not speech_presenter.get_presenter().get_only_speak_displayed_text()
        )
        image = None
        if not string:
            image = next(
                (
                    obj
                    for obj, _offset in (range_start, range_end)
                    if AXUtilities.is_image_or_canvas(obj)
                ),
                None,
            )
            if image is None:
                return False

        self._present_pending_page_change(selection_obj)
        if image is not None:
            presentation_manager.get_manager().present_object(
                script,
                image,
                generate_braille=False,
            )
            if speak_message:
                presentation_manager.get_manager().speak_message(message)
            return True

        spoken_string = string.strip() or string
        if len(string) > 5000 and speak_message:
            if message == messages.TEXT_SELECTED:
                presentation_manager.get_manager().speak_message(
                    messages.selected_character_count(len(string)),
                )
            else:
                presentation_manager.get_manager().speak_message(
                    messages.unselected_character_count(len(string)),
                )
            return True

        speech_presenter.get_presenter().speak_phrase(
            script,
            start_obj,
            start_offset,
            end_offset if start_obj == end_obj else start_offset + 1,
            spoken_string,
        )
        if speak_message:
            presentation_manager.get_manager().speak_message(message)
        return True

    def _handle_document_change(
        self,
        script: default.Script,
        obj: Atspi.Accessible,
        speak_message: bool,
    ) -> bool:
        state, old_selection, selection = (
            text_selection_manager.get_manager().update_selection_state(obj)
        )
        if state == text_selection_manager.SelectionChangeState.UNPRESENTABLE:
            msg = "TEXT SELECTION PRESENTER: Ignoring unpresentable document selection state."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        old_start, old_end = old_selection
        start, end = selection
        tokens = [
            "TEXT SELECTION PRESENTER: Document selection event from",
            obj,
            "Old boundaries:",
            old_start,
            old_end,
            "New boundaries:",
            start,
            end,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        old_start_obj, _old_start_offset = old_start
        old_end_obj, _old_end_offset = old_end
        start_obj, _start_offset = start
        end_obj, _end_offset = end
        old_elements = AXUtilities.get_text_selection_elements(old_start_obj, old_end_obj)
        new_elements = AXUtilities.get_text_selection_elements(start_obj, end_obj)
        if start == old_start and end == old_end:
            msg = "TEXT SELECTION PRESENTER: Ignoring duplicate document selection boundaries."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            for element in new_elements:
                AXUtilities.update_cached_selected_text(element)
            return bool(new_elements)

        elements = []
        for element in old_elements + new_elements:
            if element not in elements:
                elements.append(element)

        tokens = [
            "TEXT SELECTION PRESENTER: Document selection element count:",
            len(elements),
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not elements:
            return False

        if self._present_document_text_change(
            script,
            old_start,
            old_end,
            start,
            end,
            speak_message,
        ):
            for element in elements:
                AXUtilities.update_cached_selected_text(element)
            return True

        boundary_objects = (old_start_obj, old_end_obj, start_obj, end_obj)
        for element in elements:
            if element not in boundary_objects and AXUtilities.find_ancestor(
                element,
                lambda x: x in elements,
            ):
                tokens = ["TEXT SELECTION PRESENTER: Updating nested selection cache for", element]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                AXUtilities.update_cached_selected_text(element)
            else:
                tokens = ["TEXT SELECTION PRESENTER: Presenting selection change for", element]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                self._handle_basic_change(script, element, speak_message)
        return True

    def present_text_selection_change(
        self,
        script: default.Script,
        obj: Atspi.Accessible,
        speak_message: bool = True,
    ) -> bool:
        """Presents a change in selected text."""

        selection_manager = text_selection_manager.get_manager()
        managed_document = None
        if selection_manager.get_current_selection_command() is not None:
            active_document = script.utilities.active_document()
            if (
                active_document is not None
                and selection_manager.get_current_selection_command(active_document) is not None
            ):
                managed_document = active_document
        is_document = bool(
            obj is not None and (AXUtilities.is_web_element(obj) or managed_document is not None)
        )
        tokens = [
            "TEXT SELECTION PRESENTER: Presenting change for",
            obj,
            "Script:",
            script,
            "speak message:",
            speak_message,
            "is document selection:",
            is_document,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        if (
            is_document
            and script.utilities.in_document_content(obj)
            and not document_presenter.get_presenter().in_focus_mode(script.app)
        ):
            return self._handle_document_change(script, obj, speak_message)
        return self._handle_basic_change(script, obj, speak_message)


_presenter: TextSelectionPresenter = TextSelectionPresenter()


def get_presenter() -> TextSelectionPresenter:
    """Returns the text selection presenter."""

    return _presenter



# Orca
#
# Copyright 2005-2009 Sun Microsystems Inc.
# Copyright 2011-2026 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Command definitions for screen reader management."""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import cmdnames, keybindings
from .command import Command, KeyboardCommand

if TYPE_CHECKING:
    from .screen_reader_manager import ScreenReaderManager


def get_commands(owner: ScreenReaderManager) -> list[Command]:
    """Returns commands for screen reader management."""

    kb_orca_space = keybindings.KeyBinding("space", keybindings.ORCA_MODIFIER_MASK)
    kb_orca_control_space = keybindings.KeyBinding(
        "space",
        keybindings.ORCA_CTRL_MODIFIER_MASK,
    )

    return [
        KeyboardCommand(
            "shutdownHandler",
            owner.quit,
            owner.GROUP_LABEL,
            cmdnames.QUIT_ORCA,
        ),
        KeyboardCommand(
            "preferencesSettingsHandler",
            owner.show_preferences_gui,
            owner.GROUP_LABEL,
            cmdnames.SHOW_PREFERENCES_GUI,
            desktop_keybinding=kb_orca_space,
            laptop_keybinding=kb_orca_space,
        ),
        KeyboardCommand(
            "appPreferencesSettingsHandler",
            owner.show_app_preferences_gui,
            owner.GROUP_LABEL,
            cmdnames.SHOW_APP_PREFERENCES_GUI,
            desktop_keybinding=kb_orca_control_space,
            laptop_keybinding=kb_orca_control_space,
        ),
    ]
